from mock import Mock


class MockPlugin(Mock):

    """
    Provides mock Plugin object the emulates the real Plugin.

    """
    def __init__(self, *args, **kwargs):
        name = 'Unnamed'
        if 'name' in kwargs:
            name = kwargs['name']
            del kwargs['name']

        super(MockPlugin, self).__init__(*args, **kwargs)
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    def _get_child_mock(self, *args, **kwargs):
        """
        Return a normal Mock instead of a MockPlugin.
        """
        return Mock(*args, **kwargs)



from os import mkdir, stat, chmod, listdir
from os.path import join, isdir
from stat import S_IXUSR, S_IXGRP, S_IXOTH
from functools import wraps
from ConfigParser import SafeConfigParser

from jig.exc import (NotGitRepo, AlreadyInitialized,
    GitRepoNotInitialized)
from jig.conf import (JIG_DIR_NAME, JIG_PLUGIN_CONFIG_FILENAME,
    JIG_PLUGIN_DIR, PLUGIN_CONFIG_FILENAME, PLUGIN_PRE_COMMIT_SCRIPT,
    PLUGIN_PRE_COMMIT_TEMPLATE_DIR)
from jig.gitutils import is_git_repo, repo_jiginitialized
from jig.tools import slugify


def _git_check(func):
    """
    Checks to see if the directory is a Git repo.

    Raises :py:exc:`NotGitRepo` if it's not.

    You can only use this decorator on a function where the first argument is
    the path to the Git repository.
    """
    @wraps(func)
    def wrapper(gitrepo, *args, **kwargs):
        # Is it a Git repo?
        if not is_git_repo(gitrepo):
            raise NotGitRepo('Trying to initialize a directory that is not a '
                'Git repository.')

        return func(gitrepo, *args, **kwargs)
    return wrapper


@_git_check
def initializer(gitrepo):
    """
    Initializes a Git repo for use with jig.

    This will create a directory in the root of the Git repo that will contain
    files (plugins) and configuration.
    """
    # If it's already initialized, refuse to run
    if repo_jiginitialized(gitrepo):
        raise AlreadyInitialized('The repository is already initialized.')

    # Create the container for all things jig
    jig_dir = join(gitrepo, JIG_DIR_NAME)

    mkdir(jig_dir)
    mkdir(join(jig_dir, JIG_PLUGIN_DIR))

    return set_jigconfig(gitrepo)


@_git_check
def set_jigconfig(gitrepo, config=None):
    """
    Saves the config for jig in the Git repo.

    The ``config`` argument must be an instance of :py:class:`ConfigParser`.
    """
    # If it's already initialized, refuse to run
    if not repo_jiginitialized(gitrepo):
        raise GitRepoNotInitialized('The repository has not been initialized.')

    # Create the container for all things jig
    jig_dir = join(gitrepo, JIG_DIR_NAME)

    # Create an empty config parser if we were not passed one
    plugins = config if config else SafeConfigParser()

    # Create a plugin list file
    with open(join(jig_dir, JIG_PLUGIN_CONFIG_FILENAME), 'w') as fh:
        plugins.write(fh)

        return plugins


@_git_check
def get_jigconfig(gitrepo):
    """
    Gets the config for a jig initialized Git repo.
    """
    jig_dir = join(gitrepo, JIG_DIR_NAME)

    if not repo_jiginitialized(gitrepo):
        raise GitRepoNotInitialized(
            'This repository has not been initialized.')

    with open(join(jig_dir, JIG_PLUGIN_CONFIG_FILENAME), 'r') as fh:
        plugins = SafeConfigParser()
        plugins.readfp(fh)

        return plugins


def create_plugin(in_dir, bundle, name, template='python', settings={}):
    """
    Creates a plugin in the given directory.

    The directory ``in_dir`` must already exist.

    The plugin will be created with the given ``bundle`` and ``name``. These
    will be used in the plugin configuration file.

    The ``template`` specifies what scripting language will be used for the
    pre-commit executable. Existing templates can be found in the
    :file:`data/pre-commits` directory.
    """
    if not isdir(in_dir):
        raise ValueError('{} must be a directory.'.format(in_dir))

    # Create our plugin configuration
    config = SafeConfigParser()
    config.add_section('plugin')
    config.add_section('settings')
    config.set('plugin', 'bundle', bundle)
    config.set('plugin', 'name', name)

    # Add settings if applicable
    if settings:
        for key, val in settings.items():
            config.set('settings', key, str(val))

    # Create a safe directory name from the plugin name
    plugin_dir = join(in_dir, slugify(name))
    config_filename = join(plugin_dir, PLUGIN_CONFIG_FILENAME)
    pre_commit_filename = join(plugin_dir, PLUGIN_PRE_COMMIT_SCRIPT)

    # Create the directory and files
    mkdir(plugin_dir)

    with open(config_filename, 'w') as fh:
        config.write(fh)

    with open(pre_commit_filename, 'w') as fh:
        fh.write(open(join(PLUGIN_PRE_COMMIT_TEMPLATE_DIR, template)).read())

    # And make it executable
    sinfo = stat(pre_commit_filename)
    mode = sinfo.st_mode | S_IXUSR | S_IXGRP | S_IXOTH
    chmod(pre_commit_filename, mode)

    return plugin_dir


def available_templates():
    """
    Provide a list of available pre-commit templates for plugin creation.

    Templates are in :file:`jig/data`.

    This can be provided to :py:function:`create_plugin` as the
    :py:arg:`template` argument.
    """
    return listdir(PLUGIN_PRE_COMMIT_TEMPLATE_DIR)



import sys
from functools import wraps
from StringIO import StringIO
from contextlib import contextmanager

from jig.exc import ForcedExit

# Message types
INFO = u'info'
WARN = u'warn'
STOP = u'stop'


def green_bold(payload):
    """
    Format payload as green.
    """
    return u'\x1b[32;1m{}\x1b[39;22m'.format(payload)


def yellow_bold(payload):
    """
    Format payload as yellow.
    """
    return u'\x1b[33;1m{}\x1b[39;22m'.format(payload)


def red_bold(payload):
    """
    Format payload as red.
    """
    return u'\x1b[31;1m{}\x1b[39;22m'.format(payload)


def strip_paint(payload):
    """
    Removes any console specific color characters.

    Where ``payload`` is a string containing special characters used to print
    colored output to the terminal.

    Returns a unicode string without the paint.
    """
    strip = [u'\x1b[31;1m', u'\x1b[32;1m', u'\x1b[33;1m', u'\x1b[39;22m']
    for paint in strip:
        payload = payload.replace(paint, '')
    return payload


def lookup_type(strtype):
    """
    Returns the actual type for a string message representation of it.

    For example::

        >>> lookup_type('Info')
        u'info'
        >>> lookup_type('warn')
        u'warn'
        >>> lookup_type('s')
        u'stop'

    It will default to ``INFO``.

        >>> lookup_type('unknown'):
        u'info'

    """
    strtype = unicode(strtype) or u''
    mt = strtype.lower()
    if mt.startswith(u'i'):
        return INFO
    if mt.startswith(u'w'):
        return WARN
    if mt.startswith(u's'):
        return STOP

    # Default to INFO
    return INFO


class Message(object):

    """
    Represents one message that a plugin is communicating to the user.

    """
    def __init__(self, plugin, type=INFO, body='', file=None, line=None):
        """
        Create a message object associated with a plugin.

        All messages must be associated with the Plugin ``plugin`` that was
        responsible for creating them.
        """
        self.plugin = plugin

        self.type = type
        self.body = body
        self.file = file
        self.line = line

    def __repr__(self):
        reprstr = '<{cls} type="{t}", body={b}, file={f}, line={l}>'
        return reprstr.format(cls=self.__class__.__name__,
            t=self.type, b=repr(self.body), f=repr(self.file), l=self.line)

    def __eq__(self, other):
        """
        If type, body, file, and line attributes are the same they are equal.
        """
        try:
            attrs = ('type', 'body', 'file', 'line')
            for attr in attrs:
                if not getattr(self, attr) == getattr(other, attr):
                    return False
        except AttributeError:
            # If the other object is missing an attribute they can't be equal.
            return False
        return True

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = lookup_type(value)


class Error(Message):

    """
    An error message related to a plugin's results.

    """
    def __init__(self, *args, **kwargs):
        if 'type' not in kwargs:
            # Default to stop for errors
            kwargs['type'] = 'stop'

        super(Error, self).__init__(*args, **kwargs)


class ConsoleView(object):

    """
    Main view used to handle output to the console.

    """
    def __init__(self, collect_output=False, exit_on_exception=True,
            stdout=None, stderr=None):
        # Do we collect output? False means we print it out
        self.collect_output = collect_output
        self.exit_on_exception = exit_on_exception
        self._collect = {
            'stdout': stdout or StringIO(), 'stderr': stderr or StringIO()}

    @contextmanager
    def out(self):
        collected = []

        try:
            yield collected

            fo = self._collect['stdout'] if self.collect_output else sys.stdout

            for line in collected:
                fo.write(unicode(line) + u'\n')
        except Exception as e:
            fo = self._collect['stderr'] if self.collect_output else sys.stderr
            fo.write(unicode(e) + u'\n')

            try:
                retcode = e.retcode
            except AttributeError:
                # This exception does not have a return code, assume 1
                retcode = 1

            if self.exit_on_exception:
                sys.exit(retcode)   # pragma: no cover
            else:
                raise ForcedExit(retcode)

    def print_results(self, results):
        """
        Format and print plugins results.
        """
        if not results:
            return

        collater = ResultsCollater(results)

        plugins = collater.plugins
        errors = collater.errors
        reporters = collater.reporters

        form = u'plugin' if len(plugins) == 1 else u'plugins'

        if len(reporters) == 0 and len(errors) == 0:
            # Nothing to report
            with self.out() as out:
                form = u'plugin' if len(plugins) == 1 else u'plugins'
                out.append(u'Ran {plen} {form}, nothing to report'.format(
                    plen=len(plugins), form=form))
                return

        # Gather the distinct message types from the results
        cm, fm, lm = collater.messages

        # Order them from least specific to most specific, put the errors last
        messages = cm + fm + lm + errors

        # How do our message types map to a symbol
        type_to_symbol = {
            INFO: green_bold(u'\u2713'),
            WARN: yellow_bold(u'\u26a0'),
            STOP: red_bold(u'\u2715')}

        ic, wc, sc = (0, 0, 0)
        with self.out() as out:
            last_plugin = None
            for msg in messages:
                if last_plugin != msg.plugin:
                    out.append(u'\u25be  {}'.format(msg.plugin.name))
                    out.append('')
                    last_plugin = msg.plugin
                colorized = u'{}  {}'.format(
                    type_to_symbol[msg.type], self._format_message(msg))
                out.extend(colorized.splitlines())
                out.append('')

            out.append(u'Ran {plen} {form}'.format(
                plen=len(plugins), form=form))

            ic, wc, sc = [i[1] for i in collater.counts.items()]
            info = green_bold(ic) if ic else ic
            warn = yellow_bold(wc) if wc else wc
            stop = red_bold(sc) if sc else sc

            out.append(u'    Info {ic} Warn {wc} Stop {sc}'.format(
                ic=info, wc=warn, sc=stop))

            if len(errors):
                out.append(u'    ({ec} {form} reported errors)'.format(
                    ec=len(errors), form=form))

        # Return the counts for the different types of messages
        return (ic, wc, sc)

    def print_help(self, commands):
        """
        Format and print help for using the console script.
        """
        with self.out() as out:
            out.append('usage: jig [-h] COMMAND')
            out.append('')

            out.append('optional arguments:')
            out.append('  -h, --help  show this help message and exit')
            out.append('')

            out.append('jig commands:')
            for command in commands:
                name = command.__module__.split('.')[-1]
                description = command.parser.description

                out.append('  {name:12}{description}'.format(
                    name=name, description=description))

            out.append('')
            out.append('See `jig COMMAND --help` for more information')

    def _format_message(self, msg):
        """
        Formats a single message to a string.
        """
        out = []
        header = u''
        body = u''

        if msg.line:
            header += u'line {}: '.format(msg.line)

        if msg.file:
            header += msg.file

        if header:
            body = u'    {}'.format(msg.body)
        else:
            body = u'{}'.format(msg.body)

        if header:
            out.append(header)

        out.append(body)

        return '\n'.join(out)


class ResultsCollater(object):

    """
    Collects and combines plugin results into a unified summary.

    """
    def __init__(self, results):
        # Decorate our message methods
        setattr(self, '_commit_specific_message',
            self.iterresults(self._commit_specific_message))
        setattr(self, '_file_specific_message',
            self.iterresults(self._file_specific_message))
        setattr(self, '_line_specific_message',
            self.iterresults(self._line_specific_message))

        self._results = results
        self._plugins = set()
        self._reporters = set()
        self._counts = {INFO: 0, WARN: 0, STOP: 0}
        self._errors = []

        # Pre-compute our messages (collate)
        self._cm = list(self._commit_specific_message())
        self._fm = list(self._file_specific_message())
        self._lm = list(self._line_specific_message())

    @property
    def messages(self):
        """
        Messages by type for the plugin results.

        Return a tuple of messages by type based on the results that were
        provided when initializing the collater.

        Each tuple contains a generator object which will return
        ``jig.output.Message`` objects.

        The tuple has a length of 3 and is in this order:

            1. Commit specific messages
            2. File specific messages
            3. Line specific messages
        """
        return (self._cm, self._fm, self._lm)

    @property
    def plugins(self):
        """
        Provides a set of plugins that were present in the results.

        This method will return a plugin regardless of whether it yielded
        messages or not.
        """
        return self._plugins

    @property
    def reporters(self):
        """
        Provides a set of plugins that yielded messages.

        This method will only provide something other than an empty set when
        the commit, file, or line specific message methods have been called.
        """
        return self._reporters

    @property
    def counts(self):
        """
        Tally of the type of messages from the results.

        Returns a dictionary like::

            {u'info': 5, u'warn': 0, u'stop', 1}
        """
        return self._counts

    @property
    def errors(self):
        """
        Errors that were generated during collation.

        Errors are found when a piece of data given to one of the collaters is
        of a type that can't be understood.

        Returns a list of ``jig.output.Error`` objects.
        """
        return self._errors

    def iterresults(self, func):
        """
        Decorator that iterates through results.

        This simplifies some of the boilerplate for our collation. The
        decorated function must be a generator that yields ``Message`` or
        ``Error`` object. It will sift errors and collect those into a separate
        container. The ``Message`` objects then be returned to the caller.
        """
        @wraps(func)
        def wrapper(*args, **kwargs):
            for plugin, result in self._results.items():
                self._plugins.add(plugin)

                retcode, stdout, stderr = result

                if not retcode == 0:
                    error = Error(plugin)
                    error.body = stderr
                    self._errors.append(error)
                    # Remove this plugin since it's an error. If we don't do
                    # this we'll end up reporting on this 3 times.
                    del self._results[plugin]
                    continue

                for message in list(func(plugin, stdout)):
                    if isinstance(message, Error):
                        self._errors.append(message)
                        try:
                            del self._results[plugin]
                        except KeyError:
                            pass
                        continue   # pragma: no cover
                    self._reporters.add(plugin)
                    self._counts[message.type] += 1
                    yield message

        return wrapper

    def _commit_specific_message(self, plugin, obj):
        """
        Look for plugins that are reporting generic messages.

        These messages are not specific to any file or line number. They
        generally come from plugins that are inspecting the commit as a whole
        and reporting on some characteristic. A good example of this would be a
        plugin that checked to see if any modifications were made to a docs
        directory if modifications were also made to a src directory.
        Basically, a "did you write/update the docs" message.
        """
        if not obj:
            # This is falsy, there is nothing of interest here
            return

        if isinstance(obj, dict):
            # This is for file or line specific messages
            return

        if isinstance(obj, basestring):
            # Straight up message, normalize this for our loop
            obj = [obj]

        if isinstance(obj, list):
            # It's a list of [TYPE, BODY]
            for m in obj:
                if not m:
                    continue
                if isinstance(m, basestring):
                    # Straight up message, normalize this for our loop
                    yield Message(plugin, body=m)
                    continue
                if not isinstance(m, list) or len(m) != 2:
                    yield Error(plugin, body=m)
                    continue
                if not m[1]:
                    # Empty message body, this isn't useful
                    continue
                yield Message(plugin, type=m[0], body=m[1])
            # We understood this, so no need to continue
            return

        # This object is not understood
        yield Error(plugin, body=obj)

    def _file_specific_message(self, plugin, obj):
        """
        Look for plugins that are reporting file specific messages.

        These messages are specific to a file but not necessarily to a line
        number. In general they apply to a condition that is present that
        affects the whole file. An example of this would be detecting
        underscores or camel case in the filename.
        """
        if not isinstance(obj, dict):
            # This is not a file specific messages
            return

        for filename, group in obj.items():
            if isinstance(group, basestring):
                group = [group]

            if not isinstance(group, list):
                yield Error(plugin, body=group, file=filename)
                continue

            for msg in group:
                if isinstance(msg, basestring):
                    msg = [msg]

                if not isinstance(msg, list):
                    yield Error(plugin, body=msg, file=filename)
                    continue
                if len(msg) == 0:
                    # There is nothing here of interest
                    continue
                if len(msg) == 1:
                    # Should default to info type
                    if not msg[0]:
                        continue
                    yield Message(plugin, body=msg[0], file=filename)
                    continue
                if len(msg) == 2:
                    if not msg[1]:
                        continue
                    # In the format of [TYPE, BODY]
                    yield Message(plugin, body=msg[1], type=msg[0],
                        file=filename)
                    continue
                if len(msg) == 3:
                    if not msg[2]:
                        continue
                    # In the format of [LINE, TYPE, BODY]
                    if msg[0] is not None:
                        # This is line specific, skip this
                        continue
                    yield Message(plugin, body=msg[2], type=msg[1],
                        file=filename)
                    continue

                # This object is not understood
                yield Error(plugin, body=obj)

    def _line_specific_message(self, plugin, obj):
        """
        Look for plugins that are reporting line specific messages.

        For plugins wishing to identify specific lines, they use line specific
        messages. For example, you may have a JavaScript plugin that reports
        the existence of ``console.log`` on line 45. This allows the developer
        to pinpoint the problem much quicker than file or commit specific
        messages.

        There is a lack of error handling in this method. The commit and file
        specific handlers take care of error handling for us. This method gets
        to be pretty clean.
        """
        if not isinstance(obj, dict):
            # This is not a file or line specific messages
            return

        for filename, group in obj.items():
            if isinstance(group, basestring):
                group = [group]

            for msg in group:
                if isinstance(msg, basestring):
                    msg = [msg]

                if 0 <= len(msg) <= 2:
                    # There is nothing here of interest
                    continue

                if msg[0] is None:
                    # This is not line specific
                    continue
                if not msg[2]:
                    # The body is empty
                    continue
                # In the format of [LINE, TYPE, BODY]
                yield Message(plugin, body=msg[2], type=msg[1],
                    file=filename, line=msg[0])
                continue



# coding=utf-8
import json
from os import makedirs
from os.path import join, dirname
from codecs import open
from tempfile import mkdtemp
from textwrap import dedent
from collections import OrderedDict
from copy import copy

from mock import patch

from jig.exc import (ExpectationNoTests, ExpectationFileNotFound,
    ExpectationParsingError)
from jig.tests.mocks import MockPlugin
from jig.conf import CODEC
from jig.tools import NumberedDirectoriesToGit
from jig.plugins import create_plugin, Plugin
from jig.tests.testcase import JigTestCase, PluginTestCase
from jig.plugins.testrunner import (PluginTestRunner, _indent,
    InstrumentedGitDiffIndex, PluginTestReporter, get_expectations,
    Expectation, Result, SuccessResult, FailureResult,
    REPORTER_HORIZONTAL_DIVIDER)


class TestResult(PluginTestCase):

    """
    Test the base class for success and failure results.

    """
    def test_requires_correct_args(self):
        """
        TypeError if missing arguments.
        """
        with self.assertRaises(TypeError):
            Result._make(tuple())

    def test_will_make_from_iterable(self):
        """
        Class method can create new results.
        """
        result = Result._make((None, None, None, None, None))

        self.assertEqual(
            'Result(expectation=None, actual=None, plugin=None, '
                'stdin=None, stdout=None)',
            repr(result))

    def test_convert_to_dict(self):
        """
        Can convert to a dictionary.
        """
        result = Result(None, None, None, None, None)

        self.assertEqual(OrderedDict([
            ('expectation', None),
            ('actual', None),
            ('plugin', None),
            ('stdin', None),
            ('stdout', None)]),
            result._asdict())

    def test_can_replace(self):
        """
        Can replace values.
        """
        result1 = Result(None, None, None, None, None)

        result2 = result1._replace(expectation=1, actual=1)

        self.assertEqual(
            'Result(expectation=1, actual=1, plugin=None, '
                'stdin=None, stdout=None)',
            repr(result2))

    def test_replace_checks_fields(self):
        """
        Replace will not work if the keyword is incorrect.
        """
        result = Result(None, None, None, None, None)

        with self.assertRaises(ValueError):
            result._replace(not_a_field=1)

    def test_can_copy(self):
        """
        Can copy.
        """
        result1 = Result(None, None, None, None, None)

        result2 = copy(result1)

        self.assertTrue(result1 == result2)
        self.assertFalse(result1 is result2)


class TestPluginTestRunner(PluginTestCase):

    """
    Test the ability for a plugin to test itself.

    """
    def setUp(self):
        super(TestPluginTestRunner, self).setUp()

        self.timeline_iter = 0

        self.plugindir = mkdtemp()

    def add_timeline(self, plugin_dir, files):
        """
        Adds a numbered directory with the contents from files.

        ``files`` should be a list of (filename, content). An example: ::

            self.add_timeline(plugin_dir, [
                ('a.txt', 'contents of a'),
                ('b.txt', 'contents of b')])
        """
        for filename, content in files:
            path = join(plugin_dir, 'tests',
                '{0:02d}'.format(self.timeline_iter + 1), filename)
            try:
                makedirs(dirname(path))
            except OSError:
                pass
            with open(path, 'w', CODEC) as fh:
                fh.write(content)   # pragma: no branch

        # Next run of the command will start a new numbered directory
        self.timeline_iter += 1

    def add_expectation(self, plugin_dir, content):
        """
        Creates the expect.rst file inside the plugin's test directory.
        """
        try:
            makedirs(join(plugin_dir, 'tests'))
        except OSError:
            pass

        with open(join(plugin_dir, 'tests', 'expect.rst'), 'w', CODEC) as fh:
            fh.write(dedent(content))

    def test_no_tests_to_run(self):
        """
        Plugin has no tests to run.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin')

        # Since our newly created plugin has no tests, this should be empty
        with self.assertRaises(ExpectationNoTests) as ec:
            PluginTestRunner(plugin_dir)

        self.assertIn('Could not find any tests', ec.exception.message)

    def test_no_expectations(self):
        """
        We have a test timeline without any expectations.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin')

        self.add_timeline(plugin_dir, [('a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [('a.txt', 'aa\n')])

        # Since our newly created plugin has no tests, this should be empty
        with self.assertRaises(ExpectationFileNotFound) as ec:
            PluginTestRunner(plugin_dir)

        self.assertIn('Missing expectation file', ec.exception.message)

    def test_loads_tests_and_expectations(self):
        """
        Will load the timeline and expecatations when created.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin')

        self.add_timeline(plugin_dir, [('a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [('a.txt', 'aa\n')])
        self.add_expectation(plugin_dir, '''
            .. expectation::
                :from: 01
                :to: 02

                Test output''')

        ptr = PluginTestRunner(plugin_dir)

        # Just make sure that we have the expected objects
        self.assertIsInstance(ptr.timeline, NumberedDirectoriesToGit)
        self.assertIsInstance(ptr.expectations[0], Expectation)

    def test_success_result(self):
        """
        Will run the tests and detect a success result.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin')

        self.add_timeline(plugin_dir, [('a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [('a.txt', 'aa\n')])
        self.add_expectation(plugin_dir, u'''
            .. expectation::
                :from: 01
                :to: 02

                ▾  plugin

                ✓  line 1: a.txt
                    a is -

                ✓  line 1: a.txt
                    aa is +

                Ran 1 plugin
                    Info 2 Warn 0 Stop 0''')

        ptr = PluginTestRunner(plugin_dir)

        results = ptr.run()

        self.assertEqual(1, len(results))
        self.assertIsInstance(results[0], SuccessResult)
        self.assertEqual(results[0].actual.strip(),
            results[0].expectation.output.strip())
        self.assertEqual((1, 2), results[0].expectation.range)

    def test_failure_result(self):
        """
        Will run the tests and detect a failure result.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin')

        self.add_timeline(plugin_dir, [('a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [('a.txt', 'aa\n')])
        self.add_expectation(plugin_dir, u'''
            .. expectation::
                :from: 01
                :to: 02

                ▾  plugin

                ✓  Gobbildy gook

                Ran 1 plugin
                    Info 2 Warn 0 Stop 0''')

        ptr = PluginTestRunner(plugin_dir)

        results = ptr.run()

        self.assertEqual(1, len(results))
        self.assertIsInstance(results[0], FailureResult)

    def test_can_change_settings(self):
        """
        Altering the settings will be used correctly.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin',
            settings={'verbose': 'yes'})

        self.add_timeline(plugin_dir, [('a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [('a.txt', 'aa\n')])

        self.add_expectation(plugin_dir, u'''
            .. plugin-settings::

                verbose = no

            .. expectation::
                :from: 01
                :to: 02

                ▾  plugin

                ✓  a.txt
                    File has been modified

                Ran 1 plugin
                    Info 1 Warn 0 Stop 0''')

        ptr = PluginTestRunner(plugin_dir)

        results = ptr.run()

        self.assertIsInstance(results[0], SuccessResult)

    def test_non_json_stdout(self):
        """
        Still processes if the plugin returns something other than JSON data.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin')

        self.add_timeline(plugin_dir, [('a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [('a.txt', 'aa\n')])

        self.add_expectation(plugin_dir, u'''
            .. expectation::
                :from: 01
                :to: 02

                Output''')

        ptr = PluginTestRunner(plugin_dir)

        with patch.object(Plugin, 'pre_commit'):
            Plugin.pre_commit.return_value = (0, 'Non-JSON', '')

            results = ptr.run()

        self.assertResults(u'''
            ▾  plugin

            ✓  Non-JSON

            Ran 1 plugin
                Info 1 Warn 0 Stop 0
            ''', results[0].actual)

    def test_non_zero_exit_code(self):
        """
        If the exit code is non-zero, gets stderr instead.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin')

        self.add_timeline(plugin_dir, [('a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [('a.txt', 'aa\n')])

        self.add_expectation(plugin_dir, u'''
            .. expectation::
                :from: 01
                :to: 02

                Output''')

        ptr = PluginTestRunner(plugin_dir)

        with patch.object(Plugin, 'pre_commit'):
            Plugin.pre_commit.return_value = (1, '', 'Error')

            results = ptr.run()

        self.assertResults(u'''
            Exit code: 1

            Std out:
            (none)

            Std err:
            Error''', results[0].actual)

    def test_multiple_expectations(self):
        """
        Multiple tests can be ran.
        """
        plugin_dir = create_plugin(self.plugindir, 'bundle', 'plugin',
            settings={'verbose': 'no'})

        self.add_timeline(plugin_dir, [
            ('src/a.txt', 'a\n')])
        self.add_timeline(plugin_dir, [
            ('src/a.txt', 'aa\n')])
        self.add_timeline(plugin_dir, [
            ('src/a.txt', 'aaa\n')])
        self.add_timeline(plugin_dir, [
            ('src/a.txt', 'aaa\n'),
            ('src/b.txt', 'bbb\n')])

        self.add_expectation(plugin_dir, u'''
            .. expectation::
                :from: 01
                :to: 02

                ▾  plugin

                ✓  src/a.txt
                    File has been modified

                Ran 1 plugin
                    Info 1 Warn 0 Stop 0

            .. expectation::
                :from: 02
                :to: 03

                ▾  plugin

                ✓  src/a.txt
                    File has been modified

                Ran 1 plugin
                    Info 1 Warn 0 Stop 0

            .. expectation::
                :from: 03
                :to: 04

                ▾  plugin

                ✓  src/b.txt
                    File has been modified

                Ran 1 plugin
                    Info 1 Warn 0 Stop 0''')

        ptr = PluginTestRunner(plugin_dir)

        results = ptr.run()

        success = [isinstance(i, SuccessResult) for i in results]

        self.assertTrue(all(success))


class TestPluginTestReporter(PluginTestCase):

    """
    Results from a test run can be formatted for output.

    """
    def test_no_results(self):
        """
        No results reports nothing but the count.
        """
        results = []

        ptr = PluginTestReporter(results)

        self.assertEqual(u'Pass 0, Fail 0', ptr.dumps())

    def test_single_failure(self):
        """
        Reports a single failure.
        """
        expectation = Expectation((1, 2), None, u'aaa')
        results = [
            FailureResult(actual=u'bbb', expectation=expectation,
                plugin=MockPlugin())]

        ptr = PluginTestReporter(results)

        self.assertResults(u'''
            01 – 02 Fail

            Actual
            {0}

            bbb

            Diff
            {0}

            - aaa
            + bbb

            Pass 0, Fail 1'''.format(REPORTER_HORIZONTAL_DIVIDER),
            ptr.dumps())

    def test_single_success(self):
        """
        Report a single success.
        """
        expectation = Expectation((1, 2), None, u'aaa')
        results = [
            SuccessResult(actual=u'aaa', expectation=expectation,
                plugin=MockPlugin())]

        ptr = PluginTestReporter(results)

        self.assertResults(u'''
            01 – 02 Pass

            Pass 1, Fail 0''', ptr.dumps())

    def test_failure_within_success(self):
        """
        Multiple results.
        """
        expectation1 = Expectation((1, 2), None, u'aaa')
        expectation2 = Expectation((2, 3), None, u'b\nb\nb\n')
        expectation3 = Expectation((3, 4), None, u'ccc')
        results = [
            SuccessResult(actual=u'aaa', expectation=expectation1,
                plugin=MockPlugin()),
            FailureResult(actual=u'b\nB\nb\n', expectation=expectation2,
                plugin=MockPlugin()),
            SuccessResult(actual=u'ccc', expectation=expectation3,
                plugin=MockPlugin())]

        ptr = PluginTestReporter(results)

        self.assertResults(u'''
            01 – 02 Pass

            02 – 03 Fail

            Actual
            {0}

            b
            B
            b

            Diff
            {0}

              b
            + B
              b
            - b

            03 – 04 Pass

            Pass 2, Fail 1'''.format(REPORTER_HORIZONTAL_DIVIDER),
            ptr.dumps())

    def test_outputs_stdin_stdout(self):
        """
        Will report the input and output of a plugin.
        """
        stdin = json.dumps(['a', 'b', 'c'])
        stdout = json.dumps(['d', 'e', 'f'])

        expectation = Expectation((1, 2), None, u'aaa')
        results = [
            SuccessResult(actual=u'aaa', expectation=expectation,
                plugin=MockPlugin(), stdin=stdin, stdout=stdout)]

        ptr = PluginTestReporter(results)

        self.assertResults(u'''
            01 – 02 Pass

            stdin (sent to the plugin)

                [
                  "a", 
                  "b", 
                  "c"
                ]

            stdout (received from the plugin)

                [
                  "d", 
                  "e", 
                  "f"
                ]

            {0}
            Pass 1, Fail 0'''.format(REPORTER_HORIZONTAL_DIVIDER),
            ptr.dumps(verbose=True))

    def tests_outputs_stdout_not_json(self):
        """
        Will report input and output, even if it's not JSON.
        """
        stdin = 'a\nb\nc\n'
        stdout = 'd\ne\nf\n'

        expectation = Expectation((1, 2), None, u'aaa')
        results = [
            SuccessResult(actual=u'aaa', expectation=expectation,
                plugin=MockPlugin(), stdin=stdin, stdout=stdout)]

        ptr = PluginTestReporter(results)

        self.assertResults(u'''
            01 – 02 Pass

            stdin (sent to the plugin)

                a
                b
                c

            stdout (received from the plugin)

                d
                e
                f

            {0}
            Pass 1, Fail 0'''.format(REPORTER_HORIZONTAL_DIVIDER),
            ptr.dumps(verbose=True))


class TestGetExpectations(PluginTestCase):

    """
    Test converting reStructuredText documents into :py:class:`Expectation`.

    """
    def test_missing_arguments(self):
        """
        Expectation directives have two options.
        """
        with self.assertRaises(ExpectationParsingError) as ec:
            list(get_expectations(dedent('''
                .. expectation::

                    Line 1
                ''')))

        self.assertIn(
            'expectation directive requires `to` and `from` arguments',
            ec.exception.message)

    def test_non_integer_arguments(self):
        """
        Expectation directives have two required arguments.
        """
        with self.assertRaises(ExpectationParsingError) as ec:
            list(get_expectations(dedent('''
                .. expectation::
                    :to: a
                    :from: b

                    Line 1
                ''')))

        self.assertIn('Error in "expectation" directive',
            ec.exception.message)
        # And it's warning us about not being able to convert to an integer
        self.assertIn('invalid literal for int()',
            ec.exception.message)

    def test_expectation_no_settings(self):
        """
        Expectation with no accompanying settings.
        """
        exps = list(get_expectations(dedent('''
            Title 1
            =======

            .. expectation::
                :from: 01
                :to: 02

                Output''')))

        self.assertEqual(1, len(exps))
        self.assertEqual(
            Expectation(range=(1, 2), settings=None, output=u'Output'),
            exps[0])

    def test_settings(self):
        """
        Expectation with settings.
        """
        exps = list(get_expectations(dedent('''
            Title 1
            =======

            .. plugin-settings::

                a = 1
                b = 2

            .. expectation::
                :from: 01
                :to: 02

                Output''')))

        self.assertEqual(1, len(exps))
        self.assertEqual(
            Expectation((1, 2), {'a': '1', 'b': '2'}, u'Output'),
            exps[0])

    def test_settings_at_distance(self):
        """
        If settings are separated by other non structural nodes.
        """
        exps = list(get_expectations(dedent('''
            Title 1
            =======

            .. plugin-settings::

                a = 1
                b = 2

            Paragraph 1

            Paragraph 2

            .. expectation::
                :from: 01
                :to: 02

                Output''')))

        self.assertEqual(
            Expectation((1, 2), {'a': '1', 'b': '2'}, u'Output'),
            exps[0])

    def test_multiple_expectations(self):
        """
        More than one expectation.
        """
        exps = list(get_expectations(dedent('''
            Title 1
            =======

            .. plugin-settings::

                a = 1

            .. expectation::
                :from: 01
                :to: 02

                Output 1

            Title 2
            =======

            .. plugin-settings::

                a = 1

            .. expectation::
                :from: 02
                :to: 03

                Output 2''')))

        self.assertEqual([
            Expectation((1, 2), {'a': '1'}, u'Output 1'),
            Expectation((2, 3), {'a': '1'}, u'Output 2')],
            exps)


class TestInstrumentedGitDiffIndex(JigTestCase):

    """
    The instrumented GitDiffIndex provides extra features for testing.

    """
    def setUp(self):
        super(TestInstrumentedGitDiffIndex, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        self.testrepo = repo
        self.testrepodir = working_dir
        self.testdiffs = diffs

    def test_no_replacement(self):
        """
        Will not make any replacements if not configured to.
        """
        igdi = InstrumentedGitDiffIndex(self.testrepo, self.testdiffs[0])

        filenames = [i['filename'] for i in igdi.files()]

        self.assertEqual(1, len(filenames))
        self.assertEqual(
            '{}/argument.txt'.format(self.testrepodir),
            filenames[0])

    def test_will_replace(self):
        """
        Will replace part of the filename path with something else.
        """
        igdi = InstrumentedGitDiffIndex(self.testrepo, self.testdiffs[0])

        igdi.replace_path = (self.testrepodir, '/path')

        filenames = [i['filename'] for i in igdi.files()]

        self.assertEqual(1, len(filenames))
        self.assertEqual(
            '/path/argument.txt', filenames[0])


class TestIndent(JigTestCase):

    """
    The indent method will indent a sequence of strings.

    """
    def test_indent_string(self):
        """
        If the payload is a string it indents and returns a string.
        """
        self.assertEqual('    a', _indent('a'))

    def test_indents_list(self):
        """
        List payload indents each item and returns a list.
        """
        self.assertEqual(
            [u'    a', u'    b', u'    c'],
            _indent(['a', 'b', 'c']))

    def test_indents_different_by(self):
        """
        Can change the default indent of 4 to a different integer.
        """
        self.assertEqual(
            [u' a', u' b', u' c'],
            _indent(['a', 'b', 'c'], by=1))

    def test_indents_different_character(self):
        """
        Can change the character used to indent to something else.
        """
        self.assertEqual(
            [u'?a', u'?b', u'?c'],
            _indent(['a', 'b', 'c'], by=1, character='?'))



"""
Collection of :py:class:`Exception` subclasses used by jig.
"""


class JigException(Exception):

    """
    Base exception.

    """
    pass


class ForcedExit(JigException):

    """
    Raised when execution of should stop.

    This exception is useful in testing. Typically the runner is configured to
    not ``sys.exit()`` on exception so this object can be used in its place.

    """
    pass


class NotGitRepo(JigException):

    """
    A directory provided does not appear to be a Git repository.

    """
    pass


class PreCommitExists(JigException):

    """
    The :file:`pre-commit` file in the hooks directory exists.

    """
    pass


class GitRepoNotInitialized(JigException):

    """
    The given Git repository has not been initialized for usage.

    """
    pass


class GitCloneError(JigException):

    """
    Attempting to clone a Git repository failed.

    """
    pass


class AlreadyInitialized(JigException):

    """
    The given Git repository has been initialized already.

    """
    pass


class PluginError(JigException):

    """
    An error occured while working with a plugin.

    """
    pass


class CommandError(JigException):

    """
    Raise when an error occurs while using any command line tool.

    """
    pass


class ExpectationError(JigException):

    """
    Base class for all error dealing with plugin expectations.

    """
    pass


class ExpectationNoTests(ExpectationError):

    """
    Raise when no tests (numbered directories) are found for a plugin.

    """
    pass


class ExpectationFileNotFound(ExpectationError):

    """
    Raised when the expectation file should be present but is not.

    """
    pass


class ExpectationParsingError(ExpectationError):

    """
    Raised when errors are found while parsing reStructuredText expectations.

    """
    pass



import argparse
import errno
from os.path import join
from urlparse import urlparse
from uuid import uuid4 as uuid

from jig.commands.base import BaseCommand
from jig.conf import JIG_DIR_NAME, JIG_PLUGIN_DIR
from jig.exc import CommandError, ExpectationError
from jig.gitutils import clone
from jig.plugins import (get_jigconfig, set_jigconfig, PluginManager,
    create_plugin, available_templates)
from jig.plugins.testrunner import (PluginTestRunner, PluginTestReporter,
    FailureResult)

_parser = argparse.ArgumentParser(
    description='Manage this repository\'s Jig plugins',
    usage='jig plugin [-h] ACTION')

_subparsers = _parser.add_subparsers(title='actions',
    description='available commands to manage plugins')

_listparser = _subparsers.add_parser('list',
    help='list installed plugins',
    usage='jig plugin list [-h] [-r] [PATH]')
_listparser.add_argument('--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_listparser.set_defaults(subcommand='list')

_addparser = _subparsers.add_parser('add',
    help='add a plugin',
    usage='jig plugin add [-h] [-r] URL|PATH')
_addparser.add_argument('plugin',
    help='URL or path to the plugin directory')
_addparser.add_argument('--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_addparser.set_defaults(subcommand='add')

_removeparser = _subparsers.add_parser('remove',
    help='remove an installed plugin',
    usage='jig plugin remove [-h] [-r] NAME [BUNDLE]')
_removeparser.add_argument('name',
    help='Plugin name')
_removeparser.add_argument('bundle', nargs='?', default=None,
    help='Bundle name')
_removeparser.add_argument('--gitrepo', '-r', default='.', dest='path',
    help='Path to the Git repository, default current directory')
_removeparser.set_defaults(subcommand='remove')

_createparser = _subparsers.add_parser('create',
    help='create a new plugin',
    usage='jig plugin create [-h] [-l] [-d] NAME BUNDLE')
_createparser.add_argument('name',
    help='Plugin name')
_createparser.add_argument('bundle',
    help='Bundle name')
_createparser.add_argument('--language', '-l', dest='template',
    default='python', help='Scripting language: {}'.format(
        ', '.join(available_templates())))
_createparser.add_argument('--dir', '-d', default='.',
    help='Create in this directory')
_createparser.set_defaults(subcommand='create')

_testparser = _subparsers.add_parser('test',
    help='run a suite of plugin tests',
    usage='jig plugin test [-h] PLUGIN')
_testparser.add_argument('plugin', nargs='?', default='.',
    help='Path to the plugin directory')
_testparser.add_argument('--verbose', '-v',
    default=False, action='store_true',
    help='Print the input and output (stdin and stdout)')
_testparser.set_defaults(subcommand='test')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        subcommand = argv.subcommand

        # Handle the actions
        getattr(self, subcommand)(argv)

    def _bundles(self, pm):
        """
        Organize plugins by bundle name.

        Returns a dict where the key is the bundle name and the value is a list
        of all plugins that are part of that bundle.
        """
        bundles = {}

        for plugin in pm.plugins:
            if plugin.bundle not in bundles:
                bundles[plugin.bundle] = []
            bundles[plugin.bundle].append(plugin)

        return bundles

    def _plugins(self, pm):
        """
        Organize plugins by plugin name.

        Returns a dict where the key is the plugin name and the value is a list
        of all plugins that have that name.
        """
        plugins = {}

        for plugin in pm.plugins:
            if plugin.name not in plugins:
                plugins[plugin.name] = []
            plugins[plugin.name].append(plugin)

        return plugins

    def list(self, argv):
        """
        List the installed plugins.
        """
        path = argv.path

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            bundles = self._bundles(pm)

            if not bundles:
                out.append(u'No plugins installed.')
                return

            out.append(u'Installed plugins\n')

            out.append(u'{h1:<25} {h2}'.format(
                h1=u'Plugin name', h2=u'Bundle name'))

            sort_bundles = sorted(bundles.items(), key=lambda b: b[0])

            for name, plugins in sort_bundles:
                sort_plugins = sorted(plugins, key=lambda p: p.name)

                for plugin in sort_plugins:
                    out.append(u'{plugin:.<25} {name}'.format(
                        name=name, plugin=plugin.name))

    def add(self, argv):
        """
        Add a plugin.
        """
        path = argv.path
        plugin = argv.plugin

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            added = self._add_path_or_url(pm, plugin, path)

            set_jigconfig(path, pm.config)

            for p in added:
                out.append('Added plugin {} in bundle {} to the '
                    'repository.'.format(p.name, p.bundle))

    def _add_path_or_url(self, pm, plugin, gitdir):
        """
        Adds a plugin by filename or URL.

        Where ``pm`` is an instance of :py:class:`PluginManager` and ``plugin``
        is either the URL to a Git Jig plugin repository or the file name of a
        Jig plugin. The ``gitdir`` is the path to the Git repository which will
        be used to find the :file:`.jig/plugins` directory.
        """
        # If this looks like a URL we will clone it first
        url = urlparse(plugin)

        if url.scheme:
            # This is a URL, let's clone it first into .jig/plugins
            # directory.
            to_dir = join(gitdir, JIG_DIR_NAME, JIG_PLUGIN_DIR, uuid().hex)
            clone(plugin, to_dir)
            plugin = to_dir

        return pm.add(plugin)

    def remove(self, argv):
        """
        Remove a plugin.

        This method is smart enough to work with only the plugin name if it
        happens to be unique. If there is more than one plugin with the same
        name but in a different bundle it will exit with an error.
        """
        path = argv.path
        name = argv.name
        bundle = argv.bundle

        with self.out() as out:
            config = get_jigconfig(path)

            pm = PluginManager(config)

            plugins = self._plugins(pm)

            # Find the bundle if it's not specified
            if name in plugins and not bundle:
                if len(plugins[name]) > 1:
                    # There are more than one plugin by this name
                    raise CommandError('More than one plugin has the name of '
                        '{}. Use the list command to see installed '
                        'plugins.'.format(name))

                bundle = plugins[name][0].bundle

            pm.remove(bundle, name)

            set_jigconfig(path, pm.config)

            out.append('Removed plugin {}'.format(name))

    def create(self, argv):
        """
        Create a new plugin.
        """
        name = argv.name
        bundle = argv.bundle
        template = argv.template
        save_dir = argv.dir

        with self.out() as out:
            if template not in available_templates():
                raise CommandError('Language {} is not supported yet, you '
                    'can fork this project and add it though!'.format(
                        template))

            try:
                plugin_dir = create_plugin(save_dir, bundle, name,
                    template=template)

                out.append('Created plugin as {}'.format(plugin_dir))
            except OSError as ose:
                if ose.errno == errno.EEXIST:
                    # File exists
                    raise CommandError('A plugin with this name already '
                        'exists in this directory: {}.'.format(save_dir))
                # Something else, raise it again
                raise ose   # pragma: no cover

    def test(self, argv):
        """
        Run the tests for a plugin.
        """
        plugin = argv.plugin
        verbose = argv.verbose

        with self.out() as out:
            try:
                ptr = PluginTestRunner(plugin)

                results = ptr.run()

                reporter = PluginTestReporter(results)

                test_results = reporter.dumps(verbose=verbose).splitlines()

                failures = [i for i in results if isinstance(i, FailureResult)]

                if failures:
                    # Raise as an error so the status code will be non-zero
                    raise CommandError('\n'.join(test_results))

                # No failures, ok to send this to stdout
                out.extend(test_results)
            except ExpectationError as e:
                raise CommandError(e.message)



import sys


def main():
    """
    Main jig command line tool.

    Provides access to all the jig user commands.
    """
    from jig.runner import Runner

    jig = Runner()
    jig.fromconsole(sys.argv)


def test():
    """
    Run the suite of tests for jig.
    """
    import nose

    nose.main(argv=['nose'] + sys.argv[1:])


def coverage():
    """
    Create console and HTML coverage reports for the full test suite.
    """
    from coverage import coverage
    import nose

    cov = coverage(branch=True, config_file=False, source=['jig'],
        omit=['*noseplugin*', '*entrypoints*', '*testcase*'])

    cov.start()

    nose.run(argv=['nose'])

    cov.stop()

    cov.report()
    cov.html_report(directory='cover')



"""
A nose plugin to ease some testing pain.
"""
from os import listdir, mkdir
from os.path import dirname, join, realpath, isdir
from tempfile import mkdtemp
from shutil import rmtree
from subprocess import call, PIPE

from nose.plugins.base import Plugin


class TestSetupError(Exception):

    """
    Indicate that something went wrong while we were trying to setup.

    """
    pass


class TestSetup(Plugin):

    """
    Provides some testing harness for jig tests runs.

    """
    enabled = True
    name = 'jig'
    score = 500

    def __init__(self):
        super(TestSetup, self).__init__()

        testrepos = ['..'] * 3 + ['.testrepos']
        self.repo_harness_dir = join(dirname(__file__), *testrepos)

    def options(self, parser, env):
        """
        Required by Nose to add options.
        """
        parser.add_option('--unicodenazi', default=False,
            action='store_true', help='Turn unicode-nazi on')

    def configure(self, options, conf):
        """
        Required by Nose to configure the plugin.
        """
        if options.unicodenazi:
            # Turn on unicode-nazi to catch unicode/str comparisons
            __import__(unicodenazi)

    def beforeTest(self, test):
        """
        Called before each test is ran.
        """
        try:
            setattr(test.test, 'gitrepodir',
                self._create_git_repo())
        except Exception as e:
            print(e)

    def afterTest(self, test):
        """
        Called after the test ran.
        """
        rmtree(test.test.gitrepodir)

    def finalize(self, result):
        """
        Called after all tests have ran and output is collected.
        """
        for path in listdir(self.repo_harness_dir):
            rmtree(join(self.repo_harness_dir, path))

        return None

    def _create_git_repo(self):
        """
        Create an empty Git repository in the .testrepos directory.

        Returns the full path to the newly created directory.
        """
        try:
            rhd = realpath(self.repo_harness_dir)
            if not isdir(rhd):
                mkdir(rhd)
            repo = mkdtemp(dir=rhd)
        except:
            raise TestSetupError('Tried to create a directory to hold '
                'the test repositories and could not.')

        retcode = call(['git', 'init', repo],
            stdin=PIPE, stdout=PIPE, stderr=PIPE)

        if retcode != 0:
            raise TestSetupError('Could not initialize a Git repository to '
                'run tests, is Git installed?')

        return repo



# coding=utf-8
import shlex
import unittest
from os import makedirs
from os.path import join, dirname
from subprocess import check_output, STDOUT, CalledProcessError
from functools import wraps
from textwrap import dedent

from mock import patch
from git import Repo

from jig.runner import Runner
from jig.plugins import initializer
from jig.diffconvert import GitDiffIndex
from jig.tools import NumberedDirectoriesToGit, cwd_bounce
from jig.output import strip_paint, ConsoleView


def cd_gitrepo(func):
    """
    Change the current working directory to the test case's Git repository.

    This uses ``self.gitrepodir`` which is created by the
    :py:module:`jig.tests.noseplugin`.
    """
    @wraps(func)
    def wrapper(testcase, *args, **kwargs):
        with cwd_bounce(testcase.gitrepodir):
            func(testcase, *args, **kwargs)

    return wrapper


class JigTestCase(unittest.TestCase):

    """
    Base test case for all jig tests.

    """
    def setUp(self):
        self.fixturesdir = join(dirname(__file__), 'fixtures')

    def assertResults(self, expected, actual):
        """
        Assert that output matches expected argument.

        This method has some special handling intended to ease testing of the
        the output we get from :py:module:`jig.output`.

        As an example, it can be used like this::

            self.assertResults(u'''
                ▾  Plugin 1

                ✓  commit

                Ran 1 plugin
                    Info 1 Warn 0 Stop 0
                ''', self.output)

        This method will automatically dedent and strip the expected and clear
        off any console formatting characters (those that turn text colors or
        bold text).
        """
        expected = dedent(expected).strip()

        actual = strip_paint(actual.strip())

        self.assertEqual(expected, actual)

    def runcmd(self, cmd):
        """
        Takes a string and runs it, returning the output and exit code.

        Return is a tuple ``(exit_code, output_str)``.
        """
        cmd_args = shlex.split(cmd)

        try:
            output = check_output(cmd_args, stderr=STDOUT)
            retcode = 0
        except CalledProcessError as cpe:
            output = cpe.output
            retcode = cpe.returncode
        except OSError:
            output = None
            retcode = None

        return (retcode, output)

    def repo_from_fixture(self, repo_name):
        """
        Creates a ``git.Repo`` from the given fixture.

        The fixture should be a directory containing numbered directories
        suitable for creating a ``NumberedDirectoriesToGit``.

        Returns a tuple of 3 objects: repo, working_dir, diffs.
        """
        ndgit = NumberedDirectoriesToGit(
            join(self.fixturesdir, repo_name))

        repo = ndgit.repo

        return (ndgit.repo, repo.working_dir, ndgit.diffs())

    def git_diff_index(self, repo, diffs):
        """
        Retrieves the ``GitDiffIndex`` for the repository and diffs.
        """
        return GitDiffIndex(self.testrepodir, diffs)

    def create_file(self, gitrepodir, name, content):
        """
        Create or a file in the Git repository.

        The name of the file can contain directories, they will be created
        automatically.

        The directory ``gitrepodir`` represents the full path to the Git
        repository. ``name`` will be a string like ``a/b/c.txt``. ``content``
        will be written to the file.

        Return ``True`` if it complete.
        """
        try:
            makedirs(dirname(join(gitrepodir, name)))
        except OSError:
            # Directory may already exist
            pass

        with open(join(gitrepodir, name), 'w') as fh:
            fh.write(content)

        return True

    def stage(self, gitrepodir, name, content):
        """
        Create or modify a file in a Git repository and stage it in the index.

        A ``git.Index`` object will be returned.
        """
        self.create_file(gitrepodir, name, content)

        repo = Repo(gitrepodir)
        repo.index.add([name])

        return repo.index

    def stage_remove(self, gitrepodir, name):
        """
        Stage a file for removal from the Git repository.

        Where ``name`` is the path to the file.
        """
        repo = Repo(gitrepodir)
        repo.index.remove([name])

        return repo.index

    def commit(self, gitrepodir, name, content):
        """
        Create or modify a file in a Git repository and commit it.

        A ``git.Commit`` object will be returned representing the commit.

        """
        index = self.stage(gitrepodir, name, content)

        return index.commit(name)


class ViewTestCase(JigTestCase):

    """
    Access to captured output for test cases that interact with a view.

    To use this test case, the ``setUp`` method of the sub-class must set
    ``self.view`` equal to the target view.

    """
    @property
    def output(self):
        """
        Gets any output from the view that has been collected.

        Returns an empty string if the view doesn't exist or no output has
        been collected.
        """
        if not hasattr(self, 'view'):
            return ''

        return self.view._collect['stdout'].getvalue()

    @property
    def error(self):
        """
        Gets any error messages generated by the view.

        Returns an empty string if the view doesn't exist or has no output.
        """
        if not hasattr(self, 'view'):
            return ''

        return self.view._collect['stderr'].getvalue()


class RunnerTestCase(ViewTestCase):

    """
    Test case for working with :py:class:`Runner` instances

    """
    def setUp(self):
        super(RunnerTestCase, self).setUp()

        self.runner = self._init_runner()

    def _init_runner(self):
        """
        Initialize a :py:class:`Runner` and returns it.

        This will configure the runner to collect output and not exit when it
        encounters an exceptions
        """
        runner = Runner()

        # Tell the view to collect output instead of printing it
        runner.view.collect_output = True
        # Don't exit when an exception occurs so our test can continue
        runner.view.exit_on_exception = False

        # Set this up so output() and error() can read the data
        self.view = runner.view

        return runner


class PluginTestCase(JigTestCase):

    """
    Base test case for plugin tests.

    """
    def setUp(self):
        super(PluginTestCase, self).setUp()

        # Initialize the repo and grab it's config file
        self.jigconfig = initializer(self.gitrepodir)

    def _add_plugin(self, config, plugindir):
        """
        Adds the plugin to a main jig config.
        """
        section = 'plugin:test01:{}'.format(plugindir)
        config.add_section(section)
        config.set(section, 'path',
            join(self.fixturesdir, plugindir))


class CommandTestCase(ViewTestCase):

    """
    Base test case for command tests.

    """
    def run_command(self, command=None):
        """
        Run a subcommand.
        """
        with patch('jig.commands.base.create_view') as cv:
            # We hijack the create_view function so we can tell it to collect
            # output and not exit on exception.
            view = ConsoleView()

            # Collect, don't print
            view.collect_output = True
            # Don't call sys.exit() on exception
            view.exit_on_exception = False

            cv.return_value = view

            # Keep a reference to this so output() and error() will work
            self.view = view

            return self.command(shlex.split(command or ''))



# coding=utf-8
from os.path import join, dirname
from unittest import TestCase

from jig.tools import NumberedDirectoriesToGit, slugify


class TestSlugify(TestCase):

    """
    Converting strings into slugs.

    """
    def test_nothing(self):
        """
        An empty string is given.
        """
        self.assertEqual(u'', slugify(u''))

    def test_normal_ascii_string(self):
        """
        ASCII string.
        """
        self.assertEqual(u'abc-def-ghi', slugify(u'Abc & Def Ghi'))

    def test_special_characters(self):
        """
        Special characters in the string.
        """
        self.assertEqual(u'abc-def-ghi', slugify(u'Abç \u0000 Def Ghi'))


class TestNumberedDirectoriesToGit(TestCase):

    """
    Utility for converting snapshots into Git repos.

    """
    def get_nd2g(self, name):
        """
        Gets a NumberedDirectoriesToGit object.

        Where ``name`` is the basename of a directory in
        :file:`src/jig/tests/fixtures/numbereddirs`.
        """
        nd = join(dirname(__file__), 'fixtures', 'numbereddirs', name)

        return NumberedDirectoriesToGit(nd)

    def get_group(self, name):
        """
        Gets a ``git.Repo`` from the group ``name``.

        Where ``name`` is the basename of a directory in
        :file:`src/jig/tests/fixtures/numbereddirs`.
        """
        return self.get_nd2g(name).repo

    def test_bad_directory(self):
        """
        Detects bad directory.
        """
        with self.assertRaises(ValueError):
            self.get_group('bad-directory')

    def test_add_one_file(self):
        """
        Adding one file.
        """
        repo = self.get_group('group-a')

        self.assertEqual(['a.txt'],
            [i.path for i in repo.commit('HEAD^1').tree])

        # We added a second file
        self.assertEqual(['a.txt', 'b.txt'],
            [i.path for i in repo.commit('HEAD').tree])

    def test_modifying_one_file(self):
        """
        Modifying one file.
        """
        repo = self.get_group('group-b')

        # Start with one file
        self.assertEqual(['a.txt'],
            [i.path for i in repo.commit('HEAD^1').tree])

        # Same filename since it was modified
        self.assertEqual(['a.txt'],
            [i.path for i in repo.commit('HEAD').tree])

        # Should be a diff between them
        diff = repo.commit('HEAD^1').diff('HEAD')

        self.assertEqual('111\n', diff[0].a_blob.data_stream.read())
        self.assertEqual('222\n', diff[0].b_blob.data_stream.read())

    def test_remove_one_file(self):
        """
        Removing one file.
        """
        repo = self.get_group('group-c')

        diff = repo.commit('HEAD^1').diff('HEAD')

        # It's been removed
        self.assertEqual('b.txt', diff[0].a_blob.path)
        self.assertEqual(None, diff[0].b_blob)

    def test_adding_two_removing_two(self):
        """
        Adding two, removing two.
        """
        repo = self.get_group('group-d')

        diff = repo.commit('HEAD^1').diff('HEAD')

        self.assertEqual('b.txt', diff[0].a_blob.path)
        self.assertEqual(None, diff[0].b_blob)

        self.assertEqual('c.txt', diff[1].a_blob.path)
        self.assertEqual(None, diff[1].b_blob)

        self.assertEqual(None, diff[2].a_blob)
        self.assertEqual('d.txt', diff[2].b_blob.path)

        self.assertEqual(None, diff[3].a_blob)
        self.assertEqual('e.txt', diff[3].b_blob.path)

    def test_add_one_modify_one_delete_one(self):
        """
        Add one, modify one, remove one.
        """
        repo = self.get_group('group-e')

        diff = repo.commit('HEAD^1').diff('HEAD')

        # We modified a.txt
        self.assertEqual('a.txt', diff[0].a_blob.path)
        self.assertEqual('a\n', diff[0].a_blob.data_stream.read())
        self.assertEqual('aa\n', diff[0].b_blob.data_stream.read())

        # We removed b.txt
        self.assertEqual('b.txt', diff[1].a_blob.path)
        self.assertEqual(None, diff[1].b_blob)

        # And we added c.txt
        self.assertEqual(None, diff[2].a_blob)
        self.assertEqual('c.txt', diff[2].b_blob.path)

    def test_move_one_file(self):
        """
        Move one file.
        """
        repo = self.get_group('group-f')

        diff = repo.commit('HEAD^1').diff('HEAD')

        self.assertEqual('a/b.txt', diff[0].a_blob.path)
        self.assertEqual(None, diff[0].b_blob)

        self.assertEqual(None, diff[1].a_blob)
        self.assertEqual('b/b.txt', diff[1].b_blob.path)

    def test_caches_repo(self):
        """
        Calling repo twice will return the same object.
        """
        nd2g = self.get_nd2g('group-a')

        self.assertEqual(id(nd2g.repo), id(nd2g.repo))

    def test_lots_of_changes(self):
        """
        Numerous changesets.
        """
        nd2g = self.get_nd2g('group-g')

        # Make sure we have the expected 5 commits
        self.assertEqual(5, len(list(nd2g.repo.iter_commits())))

        # And 4 diffs
        self.assertEqual(4, len(nd2g.diffs()))



import argparse

from jig.commands.base import BaseCommand
from jig.gitutils import hook
from jig.plugins import initializer

_parser = argparse.ArgumentParser(
    description='Initialize a Git repository for use with Jig',
    usage='jig init [-h] [PATH]')

_parser.add_argument('path', default='.', nargs='?',
    help='Path to the Git repository')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        path = argv.path

        with self.out() as out:
            hook(path)
            initializer(path)

            out.append('Git repository has been initialized for use '
                'with Jig.')



# coding=utf-8
import json
from os.path import join
from subprocess import Popen

from mock import patch

from jig.tests.testcase import PluginTestCase
from jig.exc import PluginError
from jig.plugins import PluginManager


class TestPluginManager(PluginTestCase):

    """
    Test the plugin manager.

    """
    def test_has_empty_plugin_list(self):
        """
        A freshly initialized repo has no plugins.
        """
        pm = PluginManager(self.jigconfig)

        self.assertEqual([], pm.plugins)

    def test_has_one_plugin(self):
        """
        We can add one plugin to the main config file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')

        pm = PluginManager(self.jigconfig)

        self.assertEqual(1, len(pm.plugins))

        plugin = pm.plugins[0]

        self.assertEqual('test01', plugin.bundle)
        self.assertEqual('plugin01', plugin.name)

    def test_ignores_non_plugin_sections(self):
        """
        Other sections of the config do not get seen as a plugin.
        """
        self._add_plugin(self.jigconfig, 'plugin01')

        self.jigconfig.add_section('nonplugin')
        self.jigconfig.set('nonplugin', 'def1', '1')

        pm = PluginManager(self.jigconfig)

        self.assertEqual(1, len(pm.plugins))

    def test_add_plugin(self):
        """
        Test the add method on the plugin manager.
        """
        # Config is empty
        pm = PluginManager(self.jigconfig)

        plugin = pm.add(join(self.fixturesdir, 'plugin01'))[0]

        self.assertEqual(1, len(pm.plugins))
        self.assertTrue(pm.config.has_section('plugin:test01:plugin01'))
        self.assertEqual('plugin01', plugin.name)
        self.assertEqual('test01', plugin.bundle)

    def test_could_not_parse(self):
        """
        Tests a bad plugin config from the main config.
        """
        self._add_plugin(self.jigconfig, 'plugin02')

        with self.assertRaises(PluginError) as ec:
            PluginManager(self.jigconfig)

        self.assertIn('Could not parse config file', str(ec.exception))

    def test_contains_parsing_errors(self):
        """
        Adding a bad plugin catches parsing errors.
        """
        pm = PluginManager(self.jigconfig)

        with self.assertRaises(PluginError) as ec:
            pm.add(join(self.fixturesdir, 'plugin02'))

        self.assertIn('File contains parsing errors', str(ec.exception))

    def test_add_plugin_no_config_file(self):
        """
        Will handle a plugin that has no config file.
        """
        pm = PluginManager(self.jigconfig)

        with self.assertRaises(PluginError) as ec:
            pm.add(join(self.fixturesdir, 'plugin03'))

        self.assertIn('The plugin file', str(ec.exception))
        self.assertIn('is missing', str(ec.exception))

    def test_add_plugin_no_plugin_section(self):
        """
        Will not add a plugin with missing plugin section.
        """
        pm = PluginManager(self.jigconfig)

        with self.assertRaises(PluginError) as ec:
            pm.add(join(self.fixturesdir, 'plugin04'))

        self.assertEqual(
            'The plugin config does not contain a [plugin] section.',
            str(ec.exception))

    def test_add_plugin_no_settings_section(self):
        """
        Adds a plugin if it has no settings.
        """
        pm = PluginManager(self.jigconfig)

        pm.add(join(self.fixturesdir, 'plugin05'))

        self.assertEqual(1, len(pm.plugins))

    def test_missing_bundle_and_name(self):
        """
        Will not add a plugin if it is missing a bundle or name.
        """
        pm = PluginManager(self.jigconfig)

        with self.assertRaises(PluginError) as ec:
            pm.add(join(self.fixturesdir, 'plugin06'))

        self.assertIn('Could not find the bundle or name', str(ec.exception))

    def test_add_plugin_from_directory_of_plugins(self):
        """
        Adds all the plugins in a directory of plugins.
        """
        pm = PluginManager(self.jigconfig)

        pm.add(join(self.fixturesdir, 'plugin07'))

        self.assertEqual(2, len(pm.plugins))

    def test_cannot_add_plugin_twice(self):
        """
        After a plugin has been added, it can't be added again.
        """
        pm = PluginManager(self.jigconfig)

        with self.assertRaises(PluginError) as ec:
            pm.add(join(self.fixturesdir, 'plugin01'))
            # And the second time
            pm.add(join(self.fixturesdir, 'plugin01'))

        self.assertEqual('The plugin is already installed.',
            str(ec.exception))

    def test_remove_plugin(self):
        """
        Remove a plugin.
        """
        pm = PluginManager(self.jigconfig)

        pm.add(join(self.fixturesdir, 'plugin01'))

        self.assertTrue(pm.config.has_section('plugin:test01:plugin01'))

        pm.remove('test01', 'plugin01')

        self.assertFalse(pm.config.has_section('plugin:test01:plugin01'))
        self.assertEqual([], pm.plugins)

    def test_remove_non_existent_section(self):
        """
        Try to remove a plugin that does not exist.
        """
        pm = PluginManager(self.jigconfig)

        with self.assertRaises(PluginError) as ec:
            pm.remove('bundle', 'name')

        self.assertEqual('This plugin does not exist.',
            str(ec.exception))


class TestPlugin(PluginTestCase):

    """
    Test the individual plugins.

    """
    def setUp(self):
        super(TestPlugin, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        self.testrepo = repo
        self.testrepodir = working_dir
        self.testdiffs = diffs

    def test_new_file_pre_commit(self):
        """
        Test a new file pre-commit.
        """
        pm = PluginManager(self.jigconfig)

        pm.add(join(self.fixturesdir, 'plugin01'))
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[0])

        retcode, stdout, stderr = pm.plugins[0].pre_commit(gdi)

        data = json.loads(stdout)

        # Everything should have gone splendidly
        self.assertEqual(0, retcode)
        # Should contain our file that was added
        self.assertIn('argument.txt', data)
        # And our first line should be a warning
        self.assertEqual(
            [1, u'warn', u'The cast: is +'],
            data['argument.txt'][0])

    def test_sigpipe_error(self):
        """
        If a SIGPIPE is received, handle it without blowing up.
        """
        pm = PluginManager(self.jigconfig)

        pm.add(join(self.fixturesdir, 'plugin01'))
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[0])

        with patch.object(Popen, 'communicate'):
            ose = OSError('SIGPIPE')
            ose.errno = 32

            Popen.communicate.side_effect = ose

            retcode, stdout, stderr = pm.plugins[0].pre_commit(gdi)

        self.assertEqual(1, retcode)
        self.assertEqual('', stdout)
        self.assertEqual('Error: received SIGPIPE from the command', stderr)

    def test_oserror(self):
        """
        If a generic OSError is received, don't blow up either.
        """
        pm = PluginManager(self.jigconfig)

        pm.add(join(self.fixturesdir, 'plugin01'))
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[0])

        with patch.object(Popen, 'communicate'):
            ose = OSError('Gazoonkle was discombobulated')
            ose.errno = 1

            Popen.communicate.side_effect = ose

            retcode, stdout, stderr = pm.plugins[0].pre_commit(gdi)

        self.assertEqual(1, retcode)
        self.assertEqual('', stdout)
        self.assertEqual('Gazoonkle was discombobulated', stderr)

    def test_unicode_in_stream(self):
        """
        UTF-8 encoded streams get converted back to unicode.
        """
        pm = PluginManager(self.jigconfig)

        pm.add(join(self.fixturesdir, 'plugin01'))
        gdi = self.git_diff_index(self.testrepo, self.testdiffs[0])

        with patch.object(Popen, 'communicate'):
            # Send it encoded unicode to see if it will convert it back
            Popen.communicate.return_value = (u'å∫ç'.encode('utf-8'), '')

            retcode, stdout, stderr = pm.plugins[0].pre_commit(gdi)

        self.assertEqual(u'å∫ç', stdout)
        self.assertEqual(u'', stderr)



from os import rmdir, stat
from os.path import isfile, join
from textwrap import dedent
from tempfile import mkdtemp
from ConfigParser import ConfigParser

from jig.tests.testcase import JigTestCase, PluginTestCase
from jig.exc import (NotGitRepo, AlreadyInitialized,
    GitRepoNotInitialized)
from jig.plugins import (initializer, get_jigconfig, set_jigconfig,
    PluginManager, create_plugin, available_templates)


class TestPluginConfig(JigTestCase):

    """
    Test plugin config file handling.

    """
    def test_not_git_repo(self):
        """
        Refuses to initialize a non-Git repo.
        """
        badrepo = mkdtemp()

        with self.assertRaises(NotGitRepo):
            initializer(badrepo)

        rmdir(badrepo)

    def test_can_initialize(self):
        """
        Can initialize the Git repo.
        """
        # Should return a ConfigParser instance
        plugins = initializer(self.gitrepodir)

        self.assertTrue(isinstance(plugins, ConfigParser))

    def test_already_initialized(self):
        """
        Catches an attempt to initialize a directory that already has been.
        """
        # Initialize it once
        initializer(self.gitrepodir)

        with self.assertRaises(AlreadyInitialized):
            # Initialize the same directory again
            initializer(self.gitrepodir)

    def test_get_config_not_initialized(self):
        """
        Attempting to get a config for a directory not initialized.
        """
        with self.assertRaises(GitRepoNotInitialized):
            get_jigconfig(self.gitrepodir)

    def test_get_config(self):
        """
        Get a config.
        """
        initializer(self.gitrepodir)

        plugins = get_jigconfig(self.gitrepodir)

        self.assertTrue(isinstance(plugins, ConfigParser))
        self.assertTrue(isfile(join(self.gitrepodir, '.jig', 'plugins.cfg')))

    def test_save_config_not_initialized(self):
        """
        Raises an error if saving a config where there is not Git repo.
        """
        with self.assertRaises(GitRepoNotInitialized):
            # It hasn't been initialized at this point, this should fail
            set_jigconfig(self.gitrepodir, config=None)

    def test_save_config(self):
        """
        Can save a config.
        """
        config = initializer(self.gitrepodir)

        config.add_section('test')
        config.set('test', 'foo', 'bar')

        set_jigconfig(self.gitrepodir, config=config)


class TestCreatePlugin(PluginTestCase):

    """
    Test the plugin creation utility.

    """
    def setUp(self):
        super(TestCreatePlugin, self).setUp()

        self.plugindir = mkdtemp()

    def test_list_available_templates(self):
        """
        List available templates for creating plugins.
        """
        self.assertEqual(['python'], available_templates())

    def test_missing_directory(self):
        """
        Will not try to create a plugin in a missing directory.
        """
        # Remove this so the create_plugin function has nowhere to go.
        rmdir(self.plugindir)

        with self.assertRaises(ValueError):
            create_plugin(self.plugindir, template='python',
                bundle='test', name='plugin')

    def test_creates_plugin(self):
        """
        Can create a plugin.
        """
        plugin_dir = create_plugin(self.plugindir, template='python',
            bundle='test', name='plugin')

        pre_commit_file = join(plugin_dir, 'pre-commit')

        # We have a pre-commit file
        self.assertTrue(isfile(pre_commit_file))

        # And it's executable
        sinfo = stat(pre_commit_file)
        self.assertEqual(33261, sinfo.st_mode)

        self.assertEqual(dedent("""
            [plugin]
            bundle = test
            name = plugin

            [settings]
            """).strip(),
            open(join(plugin_dir, 'config.cfg')).read().strip())

    def test_new_plugin_compat_plugin_manager(self):
        """
        New plugins are compatible with the :py:class:`PluginManager`
        """
        plugin_dir = create_plugin(self.plugindir, template='python',
            bundle='test', name='plugin')

        pm = PluginManager(self.jigconfig)

        pm.add(plugin_dir)

        self.assertEqual(1, len(pm.plugins))
        self.assertEqual('plugin', pm.plugins[0].name)



from os import rmdir
from os.path import join, realpath
from tempfile import mkdtemp

from jig.tests.testcase import JigTestCase
from jig.exc import NotGitRepo, PreCommitExists
from jig.gitutils import hook


class TestAddingHook(JigTestCase):

    """
    Can we properly hook a Git repository

    """
    def setUp(self):
        # Where will our pre-commit file be?
        self.pc_filename = realpath(join(self.gitrepodir, '.git', 'hooks',
            'pre-commit'))

    def test_not_git_directory(self):
        """
        The directory given is not a repo.
        """
        badrepo = mkdtemp()

        try:
            with self.assertRaises(NotGitRepo):
                hook(badrepo)
        finally:
            rmdir(badrepo)

    def test_will_stop_if_pre_commit_exists(self):
        """
        Stop if .git/hooks/pre-commit exists
        """
        # Let's create a pre-commit hook, which we would use
        with open(self.pc_filename, 'w') as fh:
            fh.write('#!/bin/sh')

        with self.assertRaises(PreCommitExists):
            hook(self.gitrepodir)

    def test_successfully_hooks(self):
        """
        Creates the pre-commit hook.
        """
        pc_filename = hook(self.gitrepodir)

        self.assertEqual(self.pc_filename,
            pc_filename)

    def test_hook_runs(self):
        """
        New hook will run.
        """
        pc_filename = hook(self.gitrepodir)

        retcode, output = self.runcmd(pc_filename)

        self.assertEqual(1, retcode)
        self.assertEqual('This repository has not been initialized. Run '
            'jig init GITREPO to set it up.\n', output)



"""
Collection of utilities to deal with Git.
"""
import sys
from stat import S_IXUSR, S_IXGRP, S_IXOTH
from os import stat, chmod
from os.path import isdir, isfile, join, realpath, dirname

import git
import gitdb
import async
import smmap

from jig.exc import NotGitRepo, PreCommitExists, GitCloneError
from jig.conf import JIG_DIR_NAME

# Dependencies to make jig run
BE_CAREFUL_DIR = realpath(join(dirname(__file__), '..'))
GIT_PYTHON_DIR = realpath(join(dirname(git.__file__), '..'))
GITDB_DIR = realpath(join(dirname(gitdb.__file__), '..'))
ASYNC_DIR = realpath(join(dirname(async.__file__), '..'))
SMMAP_DIR = realpath(join(dirname(smmap.__file__), '..'))

PRE_COMMIT_HOOK_SCRIPT = \
"""#!{python_executable}
from sys import path
from os.path import dirname, join

# Make sure that we can find the directory that jig is installed
path.append('{be_careful_dir}')
path.append('{git_python_dir}')
path.append('{gitdb_dir}')
path.append('{async_dir}')
path.append('{smmap_dir}')

from jig.runner import Runner

# Start up the runner, passing in the repo directory
jig = Runner()
jig.fromhook(join(dirname(__file__), '..', '..'))
"""


def is_git_repo(gitdir):
    """
    Returns boolean whether the directory appears to be a Git directory.
    """
    return isdir(join(gitdir, '.git'))


def repo_jiginitialized(gitdir):
    """
    Returns boolean ``True`` if ``jig init GITDIR`` has been ran.
    """
    return isdir(join(gitdir, JIG_DIR_NAME))


def hook(gitdir):
    """
    Places a pre-commit hook in the given directory.

    The hook will be configured to run using the version of Python that was
    used to install jig.

    Returns the full path to the newly created post-commit hook.

    Raises :py:exc:`NotGitRepo` if the directory given is not a Git repository.
    Raises :py:exc:`PreCommitExists` if there is already a Git hook for
    pre-commit present.
    """
    if not is_git_repo(gitdir):
        raise NotGitRepo('{} is not a Git repository.'.format(
            gitdir))

    pc_filename = realpath(join(gitdir, '.git', 'hooks', 'pre-commit'))

    # Is there already a hook?
    if isfile(pc_filename):
        raise PreCommitExists('{} already exists and we will not overwrite '
            'it. If you want to use jig you\'ll have to sort this out '
            'yourself.'.format(pc_filename))

    script_kwargs = {
        'python_executable': sys.executable,
        'be_careful_dir': BE_CAREFUL_DIR,
        'git_python_dir': GIT_PYTHON_DIR,
        'gitdb_dir': GITDB_DIR,
        'async_dir': ASYNC_DIR,
        'smmap_dir': SMMAP_DIR}

    with open(pc_filename, 'w') as fh:
        fh.write(PRE_COMMIT_HOOK_SCRIPT.format(**script_kwargs))

    sinfo = stat(pc_filename)
    mode = sinfo.st_mode | S_IXUSR | S_IXGRP | S_IXOTH

    # Make sure it's executable
    chmod(pc_filename, mode)

    return pc_filename


def clone(repository, to_dir):
    """
    Clone a Git repository to a directory.

    Where ``repository`` is a string representing a path or URL to the
    repository and ``to_dir`` is where the repository will be cloned.
    """
    gitobj = git.Git()

    try:
        gitobj.execute(['git', 'clone', repository, to_dir])

        return gitobj
    except git.GitCommandError as gce:
        raise GitCloneError(str(gce))



# coding=utf-8
import sys

from mock import Mock, patch

from jig.entrypoints import main
from jig.tests.testcase import ViewTestCase, CommandTestCase
from jig.commands.base import list_commands, create_view, BaseCommand


class TestCommands(ViewTestCase):

    """
    Test the main parts of the command-line utility.

    """

    help_output = '''
        usage: jig [-h] COMMAND

        optional arguments:
          -h, --help  show this help message and exit

        jig commands:
          init        Initialize a Git repository for use with Jig
          plugin      Manage this repository's Jig plugins
          runnow      Run all plugins and show the results

        See `jig COMMAND --help` for more information'''

    def setUp(self):
        self.view = create_view()

        self.view.collect_output = True
        self.view.exit_on_exception = False

    def test_main(self):
        """
        Will output help if no arguments are given
        """
        with patch.object(sys, 'stdout') as p:
            main()

        output = ''
        for stdout_call in p.write.call_args_list:
            output += stdout_call[0][0]

        self.assertResults(self.help_output, output)

    def test_main_help(self):
        """
        Will provide help menu when ran with no arguments.
        """
        commands = list_commands()

        self.view.print_help(commands)

        self.assertResults(self.help_output, self.output)


class TestBaseCommand(CommandTestCase):

    """
    Test our base command class.

    """
    def test_abstract_process(self):
        """
        The process method is abstract.
        """
        class MissingProcessCommand(BaseCommand):
            parser = Mock()

        with self.assertRaises(NotImplementedError):
            MissingProcessCommand([])



import json
import sys
from collections import OrderedDict

from git import Repo

from jig.exc import GitRepoNotInitialized
from jig.gitutils import repo_jiginitialized
from jig.diffconvert import GitDiffIndex
from jig.plugins import get_jigconfig, PluginManager
from jig.commands import get_command, list_commands
from jig.output import ConsoleView


class Runner(object):

    """
    Runs jig in a Git repo.

    """
    def __init__(self, view=None):
        self.view = view or ConsoleView()

    def fromhook(self, gitrepo, interactive=True):
        """
        Main entry point for running, typically called from pre-commit hook.

        Where ``gitrepo`` is the file path to the Git repository.

        If ``interactive`` is True then the user will be prompted to commit or
        cancel when any messages are generated by the plugins.
        """
        results = self.results(gitrepo)

        report_counts = self.view.print_results(results)

        if interactive and report_counts and sum(report_counts):
            # Git will run a pre-commit hook with stdin pointed at /dev/null.
            # We will reconnect to the tty so that raw_input works.
            sys.stdin = open('/dev/tty')
            answer = raw_input(
                '\nCommit anyway (hit enter), or "c" to cancel the commit: ')
            if answer and answer[0].lower() == 'c':
                sys.exit(1)

        sys.exit(0)

    def fromconsole(self, argv):
        """
        Console entry point for the jig script.

        Where ``argv`` is ``sys.argv``.
        """
        # Quick copy
        argv = argv[:]
        # Our script is the first element
        argv.pop(0)

        try:
            # Next argument is the command
            command = get_command(argv.pop(0))
            command(argv)   # pragma: no cover
        except (ImportError, IndexError):
            # If it's empty
            self.view.print_help(list_commands())

    def results(self, gitrepo):
        """
        Run jig in the repository and return results.

        Results will be a dictionary where the keys will be individual plugins
        and the value the result of calling their ``pre_commit()`` methods.
        """
        self.gitrepo = gitrepo

        # Is this repository initialized to use jig on?
        with self.view.out() as out:
            if not repo_jiginitialized(self.gitrepo):
                raise GitRepoNotInitialized('This repository has not been '
                    'initialized. Run jig init GITREPO to set it up.')

        pm = PluginManager(get_jigconfig(self.gitrepo))

        # Check to make sure we have some plugins to run
        with self.view.out() as out:
            if len(pm.plugins) == 0:
                out.append('There are no plugins installed, '
                    'use jig install to add some.')
                return

            self.repo = Repo(gitrepo)

            try:
                diff = self.repo.head.commit.diff()
            except ValueError:
                # No diff on head, no commits have been written yet
                out.append('This repository is empty, jig needs at '
                    'least 1 commit to continue.')
                # Let execution continue so they *can* commit that first
                # changeset.
                return

            if len(diff) == 0:
                # There is nothing changed in this repository, no need for
                # jig to run
                out.append('No staged changes in the repository, skipping '
                    'jig.')
                return

        # Our git diff index is an object that makes working with the diff much
        # easier in the context of our plugins.
        gdi = GitDiffIndex(self.gitrepo, diff)

        # Go through the plugins and gather up the results
        results = OrderedDict()
        for plugin in pm.plugins:
            retcode, stdout, stderr = plugin.pre_commit(gdi)

            try:
                # Is it JSON data?
                data = json.loads(stdout)
            except ValueError:
                # Not JSON
                data = stdout

            results[plugin] = (retcode, data, stderr)

        return results



import argparse

from jig.commands.base import BaseCommand
from jig.runner import Runner

_parser = argparse.ArgumentParser(
    description='Run all plugins and show the results',
    usage='jig runnow [-h] [PATH]')

_parser.add_argument('path', nargs='?', default='.',
    help='Path to the Git repository')


class Command(BaseCommand):
    parser = _parser

    def process(self, argv):
        path = argv.path

        # Make the runner use our view
        runner = Runner(view=self.view)

        runner.fromhook(path, interactive=False)



from os.path import join
from tempfile import mkdtemp
from shutil import rmtree

from mock import patch
from git import Git, GitCommandError

from jig.tests.testcase import JigTestCase
from jig.exc import GitCloneError
from jig.gitutils import clone, is_git_repo


class TestClone(JigTestCase):

    """
    Git utils clone method can clone a repository.

    """
    def setUp(self):
        self.workingdir = mkdtemp()

    def tearDown(self):
        rmtree(self.workingdir)

    def test_clone_valid_repo(self):
        """
        Valid repo can be cloned.
        """
        with patch.object(Git, 'execute'):
            to_dir = join(self.workingdir, 'a')

            Git.execute.return_value = 'Cloning into X'

            gitobj = clone('http://github.com/user/repo', to_dir)

            Git.execute.assert_called_with(['git', 'clone',
                'http://github.com/user/repo', to_dir])

        self.assertIsInstance(gitobj, Git)

    def test_clone_invalid_repo(self):
        """
        Invalid repo raises error.
        """
        with patch.object(Git, 'execute'):
            to_dir = join(self.workingdir, 'a')

            Git.execute.side_effect = GitCommandError(['command'], 128,
                stderr='bad command')

            with self.assertRaises(GitCloneError) as gce:
                clone('http://github.com/user/repo', to_dir)

            self.assertIn("'command' returned exit status 128: bad command",
                gce.exception)

    def test_local_directory_clone(self):
        """
        Clones a local file-based Git repository.
        """
        to_dir = join(self.workingdir, 'a')

        clone(self.gitrepodir, to_dir)

        self.assertTrue(is_git_repo(to_dir))



# coding=utf-8
import json
from codecs import open
from os.path import join, abspath
from StringIO import StringIO
from collections import namedtuple, OrderedDict
from operator import itemgetter
from ConfigParser import SafeConfigParser

from docutils import nodes, core, io
from docutils.parsers.rst import Directive, directives

from jig.exc import (ExpectationNoTests, ExpectationFileNotFound,
    ExpectationParsingError)
from jig.conf import (CODEC, PLUGIN_EXPECTATIONS_FILENAME,
    PLUGIN_TESTS_DIRECTORY)
from jig.tools import NumberedDirectoriesToGit, cwd_bounce
from jig.diffconvert import describe_diff
from jig.output import ConsoleView, strip_paint, green_bold, red_bold
from jig.plugins import PluginManager
from jig.plugins.manager import PluginDataJSONEncoder
from jig.diffconvert import GitDiffIndex

# What docutil nodes signify a structural or sectional break
DOCUTILS_DIFFERENT_SECTION_NODES = (nodes.Root, nodes.Structural,
    nodes.Titular)

# How wide do we want the columns to be when we report test output
REPORTER_COLUMN_WIDTH = 80
# A horizontal dividing line to separate sections
REPORTER_HORIZONTAL_DIVIDER = u''.join([u'·'] * REPORTER_COLUMN_WIDTH)


def _indent(payload, by=4, character=u' '):
    """
    Indents a sequence of strings with whitespace.

    By default it will indent by 4 spaces. Change the amount of indent with
    ``by`` and the character that is used with ``character``.

    Example:

        >>> print(_indent(u'Jig', by=6, character=u'-'))
        ------Jig

    """
    return_first = False
    if isinstance(payload, (basestring)):
        payload = [payload]
        return_first = True

    indented = []
    for line in payload:
        indented.append(''.join([unicode(character)] * by) + unicode(line))

    if return_first:
        return indented[0]
    return indented


def get_expectations(input_string):
    """
    Converts a .rst document into a list of :py:class:`Expectation`.

    Expectation can be expressed in single reStructuredText documents that
    serve as documentation for the plugin and also as the tests.

    The :py:class:`PluginTestRunner` understands :py:class:`Expectation`
    objects and will run the plugin to assert that a plugin behaves as the
    expectation describes.

    A couple of custom directives allow the expectation to be expressed like
    this ::

        Will check a filename
        =====================

        The filename checker plugin will look at new files being added to a
        repository and make sure they follow a simple set of rules.

        .. plugin-settings::

            message_type = warn
            underscore_in_filenames = false
            capital_letters_in_filenames = false

        These settings will not allow underscores or capital letters in
        filenames. The ``message_type`` is set to ``warn`` which will notify
        the user that problems occur but will not prevent them from being
        committed.

        .. expectation::
           :from: 01
           :to: 02

            ▾  File name checker

            ✓  New file looks OK (matches filename rules)

            Ran 1 plugin
                Info 1 Warn 0 Stop 0
    """
    warning_stream = StringIO()
    overrides = {
        'input_encodings': 'unicode',
        'warning_stream': warning_stream}

    output, pub = core.publish_programmatically(
        source_class=io.StringInput, source=input_string,
        source_path=None,
        destination_class=io.NullOutput, destination=None,
        destination_path=None,
        reader=None, reader_name='standalone',
        parser=None, parser_name='restructuredtext',
        writer=None, writer_name='null',
        settings=None, settings_spec=None,
        settings_overrides=overrides,
        config_section=None, enable_exit_status=None)

    if warning_stream.getvalue():
        raise ExpectationParsingError(warning_stream.getvalue())

    flat_nodes = pub.writer.document.traverse()
    for i, node in enumerate(flat_nodes):
        if isinstance(node, expectations_node):
            # We have an expectation
            expectation = node
            settings = None

            # Hunt for settings
            for hunt_i in range(i - 1, 0, -1):   # pragma: no branch
                contender = flat_nodes[hunt_i]
                if isinstance(contender, plugin_settings_node):
                    # We found a plugin setting in the same area as the
                    # expectation, let's use this.
                    settings = contender.settings
                    break
                if isinstance(contender, DOCUTILS_DIFFERENT_SECTION_NODES):
                    # This is a structural element, anything past this is in a
                    # different section so we will stop.
                    break

            yield Expectation(range=expectation.range, settings=settings,
                    output=expectation.rawsource)


class Result(tuple):

    """
    Result(expectation, actual, plugin, stdin, stdout)

    """
    __slots__ = ()

    _fields = ('expectation', 'actual', 'plugin', 'stdin', 'stdout')

    def __new__(_cls, expectation, actual, plugin, stdin='', stdout=''):
        return tuple.__new__(_cls, (
            expectation, actual, plugin, stdin, stdout))

    @classmethod
    def _make(cls, iterable, new=tuple.__new__, len=len):
        """
        Make a new Result object from a sequence or iterable.
        """
        result = new(cls, iterable)
        if len(result) != 5:
            raise TypeError('Expected 5 arguments, got %d' % len(result))
        return result

    def __repr__(self):
        """
        Return a nicely formatted representation string.
        """
        reprformat = 'Result(expectation=%r, actual=%r, plugin=%r, ' + \
            'stdin=%r, stdout=%r)'
        return reprformat % self

    def _asdict(self):
        """
        Return a new OrderedDict which maps field names to their values.
        """
        return OrderedDict(zip(self._fields, self))

    def _replace(_self, **kwds):
        """
        Return a new object replacing specified fields with new values.
        """
        result = _self._make(map(kwds.pop,
            ('expectation', 'actual', 'plugin', 'stdin', 'stdout'), _self))
        if kwds:
            raise ValueError('Got unexpected field names: %r' % kwds.keys())
        return result

    def __getnewargs__(self):
        """
        Return self as a plain tuple.  Used by copy and pickle.
        """
        return tuple(self)

    expectation = property(itemgetter(0), doc='Alias for field number 0')
    actual = property(itemgetter(1), doc='Alias for field number 1')
    plugin = property(itemgetter(2), doc='Alias for field number 2')
    stdin = property(itemgetter(3), doc='Alias for field number 3')
    stdout = property(itemgetter(4), doc='Alias for field number 4')


class SuccessResult(Result):

    """
    The expectation for a single plugins tests matched its output.

    """

    def __repr__(self):   # pragma: no cover
        return '<SuccessResult from={} to={}>'.format(*self.expectation.range)


class FailureResult(Result):

    """
    The expectation for a single plugins tests does not match its output.

    """
    def __repr__(self):   # pragma: no cover
        return '<FailureResult from={} to={}>'.format(*self.expectation.range)


class InstrumentedGitDiffIndex(GitDiffIndex):

    """
    A GitDiffIndex that can be specially instrumented for testing.

    """
    def __init__(self, gitrepo, difflist):
        super(InstrumentedGitDiffIndex, self).__init__(gitrepo, difflist)

        # Allows filepaths to be modified on the fly when the
        # :py:method:`files()` method is called.
        # This should be a tuple of (REAL_PATH, REPLACEMENT_PATH)
        self.replace_path = (None, None)

    def files(self):
        real_files = super(InstrumentedGitDiffIndex, self).files()

        for f in real_files:
            if all(self.replace_path):
                f['filename'] = f['filename'].replace(
                    self.replace_path[0], self.replace_path[1])
            yield f


class PluginTestRunner(object):

    """
    Run tests to verify a plugin functions as expected.

    """
    def __init__(self, plugin_dir):
        self.plugin_dir = plugin_dir
        self.timeline = None
        self.expectations = None

        try:
            test_directory = join(plugin_dir, PLUGIN_TESTS_DIRECTORY)
            self.timeline = NumberedDirectoriesToGit(test_directory)
        except ValueError:
            raise ExpectationNoTests('Could not find any tests: {}.'.format(
                test_directory))

        try:
            expect_filename = join(plugin_dir, PLUGIN_TESTS_DIRECTORY,
                PLUGIN_EXPECTATIONS_FILENAME)

            with open(expect_filename, 'r', CODEC) as fh:
                expectation_text = fh.read()   # pragma: no branch

            self.expectations = list(get_expectations(expectation_text))
        except (IOError, OSError):
            raise ExpectationFileNotFound(
                'Missing expectation file: {}.'.format(expect_filename))

    def run(self):
        """
        Run the tests for this plugin.

        Returns a list of :py:class:`Result` objects which represent the
        results from the test run.
        """
        # Use an empty config, we are not going to save this to disk
        pm = PluginManager(SafeConfigParser())

        # Add the plugin we are testing
        pm.add(self.plugin_dir)

        # The instance of our plugin we will run the pre_commit test on
        plugin = pm.plugins[0]

        # Capture the default plugin config for resets while testing
        default_settings = plugin.config

        results = []

        for exp in self.expectations:
            # Make sure that the range is off by 1
            assert exp.range[1] == exp.range[0] + 1

            # Update the plugin config (settings) if available
            if exp.settings:
                plugin.config = exp.settings
            else:
                plugin.config = default_settings

            # View to help us create the output
            view = ConsoleView(collect_output=True, exit_on_exception=False)

            # Get a GitDiffIndex object from
            gdi = InstrumentedGitDiffIndex(self.timeline.repo.working_dir,
                self.timeline.diffs()[exp.range[0] - 1])

            # What is the numbered test directory reprsenting our commit?
            wd = abspath(join(self.plugin_dir, PLUGIN_TESTS_DIRECTORY,
                '{0:02d}'.format(exp.range[1])))

            with cwd_bounce(wd):
                # Patch up the filename to be within our numbered directory
                # instead of the Git repository
                gdi.replace_path = (self.timeline.repo.working_dir, wd)

                # Gather up the input to the plugin for logging
                stdin = json.dumps({'config': plugin.config,
                    'files': gdi}, indent=2, cls=PluginDataJSONEncoder)

                # Now run the actual pre_commit hook for this plugin
                res = plugin.pre_commit(gdi)
                # Break apart into its pieces
                retcode, stdout, stderr = res   # pragma: no branch

            try:
                # Is it JSON data?
                data = json.loads(stdout)
            except ValueError:
                # Not JSON
                data = stdout

            if retcode == 0:
                # Format the results according to what you normally see in the
                # console.
                view.print_results({plugin: (retcode, data, stderr)})
            else:
                results.append(FailureResult(exp,
                    'Exit code: {}\n\nStd out:\n{}\n\nStd err:\n{}'.format(
                        retcode, stdout or '(none)', stderr or '(none)'),
                    plugin))
                continue

            # Now remove the color character sequences to make things a little
            # easier to read, copy, and paste.
            actual = strip_paint(view._collect['stdout'].getvalue() or
                view._collect['stderr'].getvalue())

            resargs = (exp, actual, plugin, stdin, stdout)
            if actual.strip() != exp.output.strip():
                results.append(FailureResult(*resargs))
            else:
                results.append(SuccessResult(*resargs))

        return results


class PluginTestReporter(object):

    """
    Formats a list of test results into human-readable format.

    The list must contain :py:class:`SuccessResult` or
    :py:class:`FailureResult' objects.

    """
    def __init__(self, results):
        self.results = results

    def _add_verbosity(self, out, stdin, stdout):
        """
        Formats the stdin and stdout into the output.
        """
        # Try to pretty print our data if it's JSON
        data = [stdin, stdout]
        for i in (0, 1):
            try:
                obj = json.loads(data[i])
                data[i] = json.dumps(obj, indent=2)
            except ValueError:
                # Wasn't JSON
                pass
            data[i] = _indent(data[i].splitlines())

        out.append(u'stdin (sent to the plugin)')
        out.append(u'')
        out.extend(data[0])
        out.append(u'')
        out.append(u'stdout (received from the plugin)')
        out.append(u'')
        out.extend(data[1])
        out.append(u'')
        out.append(REPORTER_HORIZONTAL_DIVIDER)

    def dumps(self, verbose=False):
        """
        Formats a list of test results to unicode.

        The reporter can also output the value of what was sent to stdin and
        what the plugin sent to stdout if ``verbose`` is ``True``.
        """
        out = []

        results = self.results

        for result in results:
            exprange = result.expectation.range
            v_out = []

            if verbose:
                self._add_verbosity(v_out, result.stdin, result.stdout)

            if isinstance(result, SuccessResult):
                out.append(green_bold(u'{0:02d} – {1:02d} Pass'.format(
                    exprange[0], exprange[1])))
                out.append(u'')
                out.extend(v_out)
                continue

            out.append(red_bold(u'{0:02d} – {1:02d} Fail'.format(
                exprange[0], exprange[1])))

            out.extend(v_out)

            out.append(u'')

            out.append(u'Actual')
            out.append(REPORTER_HORIZONTAL_DIVIDER)
            out.append(u'')
            out.extend(result.actual.splitlines())
            out.append(u'')

            out.append(u'Diff')
            out.append(REPORTER_HORIZONTAL_DIVIDER)
            out.append(u'')

            diff = describe_diff(result.expectation.output, result.actual)
            for (_, diff_type, line) in diff:
                if diff_type == '-':
                    decorator = red_bold
                elif diff_type == '+':
                    decorator = green_bold
                else:
                    # No operation but return
                    decorator = lambda a: a

                out.append(decorator(u'{} {}'.format(diff_type, line)))

            out.append(u'')

        pass_count = len([i for i in results if isinstance(i, SuccessResult)])
        fail_count = len([i for i in results if isinstance(i, FailureResult)])

        out.append(u'Pass {}, Fail {}'.format(pass_count, fail_count))

        return u'\n'.join(out)


Expectation = namedtuple('Expectation', 'range settings output')


class plugin_settings_node(nodes.literal_block):

    """
    Represents a docutils node specific to plugin settings.

    """
    pass


class expectations_node(nodes.literal_block):

    """
    Represents the desired output from a plugin when tested.

    """
    pass


class PluginSettingsDirective(Directive):

    """
    Docutils directive for expressing plugin settings.

    Example::

        .. plugin-settings::

            underscore_in_filenames = no
            capital_letters_in_filenames = no
    """

    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {}

    def run(self):
        code = u'\n'.join(self.content)

        # Use the config parser to get our settings
        config_fp = StringIO('[settings]\n{}'.format(code))
        config = SafeConfigParser()
        config.readfp(config_fp)
        node = plugin_settings_node(code, code)
        node.settings = dict(config.items('settings'))
        return [node]


class ExpectationDirective(Directive):

    """
    Docutils directive for documenting plugin settings.

    Example::

        .. expectation::
            :from: 01
            :to: 02

            ▾  File name checker

            ✓  New file looks OK (matches filename rules)

            Ran 1 plugin
                Info 1 Warn 0 Stop 0
    """

    has_content = True
    required_arguments = 0
    optional_arguments = 2
    final_argument_whitespace = True
    option_spec = {
        'from': directives.nonnegative_int,
        'to': directives.nonnegative_int}

    def run(self):
        code = u'\n'.join(self.content)
        node = expectations_node(code, code)

        # The from and to are required
        node.range = (self.options.get('from', None),
            self.options.get('to', None))

        if not node.range[0] or not node.range[1]:
            # The range is incomplete
            self.state_machine.reporter.error('expectation directive requires '
                '`to` and `from` arguments')

        return [node]


directives.register_directive('plugin-settings', PluginSettingsDirective)
directives.register_directive('expectation', ExpectationDirective)



from shutil import rmtree
from os.path import join
from contextlib import nested

from mock import patch

from jig.tests.testcase import RunnerTestCase, PluginTestCase
from jig.tests.mocks import MockPlugin
from jig.exc import ForcedExit
from jig.plugins import set_jigconfig, Plugin
from jig.runner import Runner


class TestRunnerEntryPoints(RunnerTestCase, PluginTestCase):

    """
    Runner hook is used by the command line script to run jig.

    """
    def setUp(self):
        super(TestRunnerEntryPoints, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        self.testrepo = repo
        self.testrepodir = working_dir
        self.testdiffs = diffs

    def test_simple_integration(self):
        """
        Will run the hook.
        """
        with patch.object(self.runner, 'results'):
            plugin = MockPlugin()
            # Empty results
            self.runner.results.return_value = {plugin:
                (0, '', '')}

            with self.assertRaises(SystemExit) as ec:
                self.runner.fromhook(self.gitrepodir)  # pragma: no branch

            self.assertEqual(0, ec.exception.code)

            self.assertEqual(
                'Ran 1 plugin, nothing to report\n',
                self.output)

    def test_will_not_prompt_if_no_messages(self):
        """
        No prompt is presented if there are no messages to communicate.
        """
        with patch.object(Runner, 'results'):
            # No results came back from any plugin
            Runner.results.return_value = []

            with nested(
                patch('jig.runner.raw_input', create=True),
                patch('jig.runner.sys'),
                self.assertRaises(SystemExit)
            ) as (ri, r_sys, ec):
                r_sys.exit.side_effect = SystemExit
                self.runner.fromhook(self.gitrepodir)

        # Make sure that the call to raw_input never happened
        self.assertFalse(ri.called)

        # And we exited with 0 because there is no reason to stop the commit
        r_sys.exit.assert_called_once_with(0)

    def test_will_prompt_user(self):
        """
        User sees a prompt if there are messages.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged changes
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.raw_input', create=True),
            patch('jig.runner.sys'),
            self.assertRaises(SystemExit)
        ) as (ri, r_sys, ec):
            # Fake the raw_input call to return 'c'
            r_sys.exit.side_effect = SystemExit
            ri.return_value = 'c'

            self.runner.fromhook(self.gitrepodir)

        # The user was prompted about committing or canceling
        ri.assert_called_once_with('\nCommit anyway (hit enter), '
            'or "c" to cancel the commit: ')
        # When they said cancel we exited with non-zero
        r_sys.exit.assert_called_once_with(1)

    def test_will_prompt_but_continue_anyway(self):
        """
        The user can choose to continue with the commit anyway.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Create staged changes
        self.commit(self.gitrepodir, 'a.txt', 'a')
        self.stage(self.gitrepodir, 'b.txt', 'b')

        with nested(
            patch('jig.runner.raw_input', create=True),
            patch('jig.runner.sys')
        ) as (ri, r_sys):
            # Fake the raw_input call to return 'c'
            ri.return_value = ''

            self.runner.fromhook(self.gitrepodir)

        # The user was prompted about committing or canceling
        ri.assert_called_once_with('\nCommit anyway (hit enter), '
            'or "c" to cancel the commit: ')
        # When they said cancel we exited with non-zero
        r_sys.exit.assert_called_once_with(0)


class TestRunnerResults(RunnerTestCase, PluginTestCase):

    """
    Can the runner iterate over the plugins and gather results.

    """
    def setUp(self):
        super(TestRunnerResults, self).setUp()

        repo, working_dir, diffs = self.repo_from_fixture('repo01')

        self.testrepo = repo
        self.testrepodir = working_dir
        self.testdiffs = diffs

    def test_uninitialized_repo(self):
        """
        The .jig directory has not been initialized.
        """
        # Remove the .jig directory, effectively un-initializing our repository
        rmtree(join(self.gitrepodir, '.jig'))

        with self.assertRaises(ForcedExit) as ec:
            self.runner.results(self.gitrepodir)

        self.assertEqual(1, ec.exception.message)
        self.assertEqual('This repository has not been initialized. Run '
            'jig init GITREPO to set it up.\n',
            self.error)

    def test_no_plugins(self):
        """
        If there is a .jig directory without any plugins.
        """
        self.runner.results(self.gitrepodir)

        self.assertEqual('There are no plugins installed, use jig '
            'install to add some.\n',
            self.output)

    def test_empty_repository(self):
        """
        If .jig is ran on a repository that hasn't had any commits
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        self.runner.results(self.gitrepodir)

        self.assertEqual('This repository is empty, jig needs at '
            'least 1 commit to continue.\n',
            self.output)

    def test_no_diff(self):
        """
        If .jig is ran on a repository without any changes.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        self.commit(self.gitrepodir,
            name='a.txt',
            content='a')

        self.runner.results(self.gitrepodir)

        self.assertEqual('No staged changes in the repository, skipping '
            'jig.\n',
            self.output)

    def test_unstaged_one_file(self):
        """
        Ran on a repository with an unstaged file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Add the first commit because we have to have it
        self.commit(self.gitrepodir,
            name='a.txt',
            content='a')

        # We've created this but not added it to the index
        self.create_file(self.gitrepodir,
            name='b.txt',
            content='b')

        self.runner.results(self.gitrepodir)

        self.assertEqual('No staged changes in the repository, skipping '
            'jig.\n',
            self.output)

    def test_staged_one_file(self):
        """
        Ran on a repository with a staged file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Add the first commit because we have to have it
        self.commit(self.gitrepodir,
            name='a.txt',
            content='a')

        # Create a new file an stage it to the index
        self.stage(self.gitrepodir,
            name='b.txt',
            content='b')

        results = self.runner.results(self.gitrepodir)

        # One plugin ran, we should have results from it
        self.assertEqual(1, len(results))

        # It's plugin01
        self.assertEqual('plugin01', results.keys()[0].name)

        retcode, stdout, stderr = results.items()[0][1]

        # The return code is 0
        self.assertEqual(0, retcode)
        # We auto-convert to an object
        self.assertEqual({u'b.txt': [[1, u'warn', u'b is +']]}, stdout)
        # And no errors occurred here
        self.assertEqual('', stderr)

    def test_modified_one_file(self):
        """
        One modified and staged file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        # Add the first commit because we have to have it
        self.commit(self.gitrepodir,
            name='a.txt',
            content='a')

        # We've created this but not added it to the index
        self.stage(self.gitrepodir,
            name='a.txt',
            content='aaa')

        results = self.runner.results(self.gitrepodir)

        _, stdout, _ = results.items()[0][1]

        self.assertEqual({u'a.txt': [[1, u'warn', u'a is -'], [1, u'warn',
            u'aaa is +']]}, stdout)

    def test_deleted_one_file(self):
        """
        Delete one file.
        """
        self._add_plugin(self.jigconfig, 'plugin01')
        set_jigconfig(self.gitrepodir, config=self.jigconfig)

        self.commit(self.gitrepodir,
            name='a.txt',
            content='a')

        # Now stage a file for removal
        self.stage_remove(self.gitrepodir, name='a.txt')

        results = self.runner.results(self.gitrepodir)

        _, stdout, _ = results.items()[0][1]

        # We should see it being removed
        self.assertEqual({u'a.txt': [[1, u'warn', u'a is -']]}, stdout)

    def test_handles_non_json_stdout(self):
        """
        Supports non-JSON output from the plugin.
        """
        with patch.object(Plugin, 'pre_commit'):
            Plugin.pre_commit.return_value = (
                0, 'Test non-JSON output', '')

            self._add_plugin(self.jigconfig, 'plugin01')
            set_jigconfig(self.gitrepodir, config=self.jigconfig)

            self.commit(self.gitrepodir,
                name='a.txt',
                content='a')

            self.stage(self.gitrepodir,
                name='b.txt',
                content='b')

            results = self.runner.results(self.gitrepodir)

            _, stdout, _ = results.items()[0][1]

        # And we can still get the output even though it's not JSON
        self.assertEqual('Test non-JSON output',
            stdout)

    def test_handles_retcode_1_with_stderr(self):
        """
        Handles non-zero return codes and data written to stderr.
        """
        with patch.object(Plugin, 'pre_commit'):
            Plugin.pre_commit.return_value = (
                1, '', 'Something went horribly wrong')

            self._add_plugin(self.jigconfig, 'plugin01')
            set_jigconfig(self.gitrepodir, config=self.jigconfig)

            self.commit(self.gitrepodir,
                name='a.txt',
                content='a')

            self.stage(self.gitrepodir,
                name='b.txt',
                content='b')

            results = self.runner.results(self.gitrepodir)

            retcode, _, stderr = results.items()[0][1]

        self.assertEqual(1, retcode)
        self.assertEqual('Something went horribly wrong',
            stderr)



from os import listdir
from os.path import join, dirname, isdir

from jig.output import ConsoleView

_commands_dir = dirname(__file__)


def list_commands():
    """
    List the commands available.
    """
    commands = []
    for f in listdir(_commands_dir):
        if isdir(join(_commands_dir, f)):
            continue
        if not f.endswith('.py'):
            continue
        try:
            commands.append(get_command(f.split('.')[0]))
        except (ImportError, AttributeError):
            continue
    return commands


def get_command(name):
    """
    Gets an instance of the named jig sub-command.

    For example::

        >>> get_command('init')
        <jig.commands.init.Command object at 0x10048fed0>
    """
    mod = __import__('jig.commands.{}'.format(name.lower()),
        globals(), locals(), ['Command'], 0)
    return mod.Command


def create_view():
    """
    Creates a view the command can use to output data.

    This method is separated from the :py:class:`BaseCommand` class to
    facilitate testing. By mocking out this method, commands can use views that
    have been configured to collect output instead of sending it to the
    terminal.
    """
    return ConsoleView()


class BaseCommand(object):

    """
    Base command for implementing script sub-commands.

    """
    def __init__(self, argv):
        """
        Parse the command line arguments and call process with the results.

        Where argv is a split string. See :py:module:`shlex`.
        """
        args = self.parser.parse_args(argv)

        # Setup something our command can use to send output
        self.view = create_view()
        # A shorter alias to the view's out decorator
        self.out = self.view.out

        # Finally, process the arguments
        self.process(args)

    def process(self, args):
        """
        Perform whatever operation this command is supposed to do.
        """
        raise NotImplementedError



import json
from os import listdir
from os.path import join, isfile, isdir, realpath
from subprocess import Popen, PIPE
from ConfigParser import SafeConfigParser
from ConfigParser import Error as ConfigParserError
from ConfigParser import NoSectionError
from collections import OrderedDict

from jig.exc import PluginError
from jig.conf import PLUGIN_CONFIG_FILENAME, PLUGIN_PRE_COMMIT_SCRIPT


class PluginManager(object):

    """
    Provides access to running and managing plugins.

    """
    def __init__(self, config):
        """
        Create a plugin manager with the given ``config``.

        The ``config`` argument should be an instance of
        :py:class:`SafeConfigParser` and will be the main configuration for an
        jig-initialized Git repository.
        """
        # The instance of SafeConfigParser we get from :py:method:`config`.
        self.config = config

        # Look through the config and initialize any installed plugins
        self._plugins = self._init_plugins(config)

    def _init_plugins(self, config):
        """
        Creates :py:class:`Plugin` instances from ``config``.
        """
        plugins = []
        for section_name in config.sections():
            if not section_name.startswith('plugin:'):
                # We are only interested in the plugin configs
                continue

            _, bundle, name = section_name.split(':')

            path = config.get(section_name, 'path')

            plugin_cfg = join(path, PLUGIN_CONFIG_FILENAME)

            with open(plugin_cfg) as fh:
                plugin_config = SafeConfigParser()
                try:
                    plugin_config.readfp(fh)   # pragma: no branch
                except ConfigParserError:
                    # Something happened when parsing the config
                    raise PluginError('Could not parse config file for '
                        '{} in {}.'.format(name, path))

            # Get rid of the path, we don't need to send this as part of the
            # config for the plugin
            pc = OrderedDict(config.items(section_name))
            del pc['path']

            section = Plugin(bundle, name, path, pc)
            plugins.append(section)

        return plugins

    def __iter__(self):
        return iter(self._plugins)

    def __len__(self):
        return len(self._plugins)

    @property
    def plugins(self):
        return list(self)

    def add(self, plugindir):
        """
        Add the given plugin or directory of plugins to this manager instance.

        ``plugindir`` should be the full path to a directory containing all the
        files required for a Jig plugin. It can alternatively be a directory of
        plugins, where each sub-directory is a Jig plugin.

        If ``recursive`` is True, then add will treat this as a directory of
        plugins instead of a single plugin and attempt to add them all.

        Returns a list of plugins that were added to this manager.
        """
        exc_collection = []
        added = []

        try:
            # Add as if plugindir is the actual plugin
            added.append(self._add_plugin(plugindir))

            return added
        except PluginError as pe:
            exc_collection.append(pe)

        # Walk the directory, try to add each sub-directory as a plugin
        for dirname in listdir(plugindir):
            subdir = join(plugindir, dirname)
            if not isdir(subdir):
                continue
            try:
                added.append(self._add_plugin(subdir))
            except PluginError as pe:
                exc_collection.append(pe)

        if added:
            return added

        # If we haven't added any plugins and we have an exception raise it
        raise exc_collection[0]

    def _add_plugin(self, plugindir):
        """
        If this is a Jig plugin, add it.

        ``plugindir`` should be the full path to a directory containing all the
        files required for a jig plugin.
        """
        # Is this a plugins?
        config_filename = join(plugindir, PLUGIN_CONFIG_FILENAME)

        if not isfile(config_filename):

            raise PluginError('The plugin file {} is missing.'.format(
                config_filename))

        config = SafeConfigParser()

        with open(config_filename, 'r') as fh:
            try:
                config.readfp(fh)
            except ConfigParserError as e:
                raise PluginError(e)

        try:
            settings = OrderedDict(config.items('settings'))
        except NoSectionError:
            settings = []

        try:
            plugin_info = OrderedDict(config.items('plugin'))
        except NoSectionError:
            raise PluginError('The plugin config does not contain a '
                '[plugin] section.')

        try:
            bundle = plugin_info['bundle']
            name = plugin_info['name']
        except KeyError:
            raise PluginError('Could not find the bundle or name of '
                'the plugin.')

        new_section = 'plugin:{bundle}:{name}'.format(
            bundle=bundle, name=name)

        if self.config.has_section(new_section):
            raise PluginError('The plugin is already installed.')

        self.config.add_section(new_section)

        self.config.set(new_section, 'path', plugindir)

        for setting in settings:
            option, value = setting, settings[setting]
            self.config.set(new_section, option, value)

        # Re-initialize the self.plugins list
        self._plugins = self._init_plugins(self.config)

        # And return the plugin once we find it
        for plugin in self._plugins:   # pragma: no branch
            if plugin.name == name and plugin.bundle == bundle:
                return plugin

    def remove(self, bundle, name):
        """
        Remove a plugin from the list and config.

        Both ``bundle`` and ``name`` are required. A
        :py:exception:`PluginError` will be raised if the plugin does not
        exist.
        """
        section_name = 'plugin:{bundle}:{name}'.format(
            bundle=bundle, name=name)

        if not self.config.has_section(section_name):
            raise PluginError('This plugin does not exist.')

        self.config.remove_section(section_name)

        # Again, re-initialize the self.plugins list
        self._plugins = self._init_plugins(self.config)


class Plugin(object):

    """
    A single unit that performs some helpful operation for the user.

    """
    def __init__(self, bundle, name, path, config={}):
        # What bundle is this plugin a part of
        self.bundle = bundle
        # What is the name of this plugin?
        self.name = name
        # Where does this plugin live
        self.path = realpath(path)
        # Plugin-specific configuration
        self.config = config

    def pre_commit(self, git_diff_index):
        """
        Runs the plugin's pre-commit script, passing in the diff.

        ``git_diff_index`` is a :py:class:`jig.diffconvert.GitDiffIndex`
        object.

        The pre-commit script will receive JSON data as standard input (stdin).
        The JSON data is comprised of two main attributes: config and diff.

        The ``config`` attribute represents the configuration for this plugin.
        This is up to the plugin author but the values can be changed by the
        user.

        The ``diff`` attribute is a list of files and changes that have
        occurred to them.  See :py:module:`jig.diffconvert` for
        information on what this object provides.
        """
        # Grab this plugin's settings
        data_in = {
            'config': self.config,
            'files': git_diff_index}

        script = join(self.path, PLUGIN_PRE_COMMIT_SCRIPT)
        ph = Popen([script], stdin=PIPE, stdout=PIPE, stderr=PIPE)

        # Send the data to the script
        stdin = json.dumps(data_in, indent=2, cls=PluginDataJSONEncoder)

        retcode = None
        stdout = ''
        stderr = ''

        try:
            stdout, stderr = ph.communicate(stdin)

            # Convert to unicode
            stdout = stdout.decode('utf-8')
            stderr = stderr.decode('utf-8')

            retcode = ph.returncode
        except OSError as ose:
            # Generic non-zero retcode that indicates an error
            retcode = 1
            if ose.errno == 32:
                stderr = u'Error: received SIGPIPE from the command'
            else:
                stderr = unicode(ose)

        # And return the relevant stuff
        return retcode, stdout, stderr


class PluginDataJSONEncoder(json.JSONEncoder):

    """
    Converts the special data objects used when a plugin runs pre-commit.

    """
    def default(self, obj):
        """
        Implements JSONEncoder default method.
        """
        files = [i for i in obj.files()]

        obj = []
        for f in files:
            obj.append({
                'type': unicode(f['type']),
                'name': unicode(f['name']),
                'filename': unicode(f['filename']),
                'diff': [j for j in f['diff']]})

        return obj
